from __future__ import annotations
from enum import IntEnum
from typing import Any, List, Optional, TypedDict, SupportsInt, Union

from discord import PartialEmoji


__all__ = (
    'OptionType',
    'ComponentType',
    'ApplicationCommandType',
    'ButtonStyle',
)


class OptionType(IntEnum):
    """
    Attributes
    ----------
    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8
    MENTIONABLE = 9
    NUMBER = 10
    """

    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8
    MENTIONABLE = 9
    NUMBER = 10


class OptionChoicePayload(TypedDict):
    name: str
    value: Any


class OptionPayload(TypedDict, total=False):
    name: str
    description: Optional[str]
    required: bool
    type: int
    choices: List[OptionChoicePayload]
    options: List[OptionPayload]  # type: ignore


class ApplicationCommandType(IntEnum):
    CHAT_INPUT = 1
    SLASH = 1
    USER = 2
    MESSAGE = 3


class ApplicationCommandPayload(TypedDict, total=False):
    id: Optional[SupportsInt]
    name: str
    type: ApplicationCommandType
    application_id: Optional[SupportsInt]


class SlashCommandPayload(ApplicationCommandPayload, total=False):
    description: str
    options: Optional[List[OptionPayload]]
    default_permission: bool


class RawCommandPermissionPayload(TypedDict):
    id: int
    type: int
    permission: bool


class ApplicationCommandPermissionsPayload(TypedDict):
    permissions: List[RawCommandPermissionPayload]


class SelectOptionPayload(TypedDict, total=False):
    label: str
    value: str
    description: Optional[str]
    emoji: Optional[Union[str, PartialEmoji]]
    default: bool


class ComponentType(IntEnum):
    """
    An enumerator for component types.

    Attributes
    ----------
    ActionRow = 1
    Button = 2
    SelectMenu = 3
    """

    ActionRow = 1
    Button = 2
    SelectMenu = 3


class ComponentPayload(TypedDict, total=False):
    disabled: bool
    custom_id: Optional[str]
    type: int


class SelectMenuPayload(ComponentPayload, total=False):
    placeholder: Optional[str]
    min_values: int
    max_values: int
    options: List[SelectOptionPayload]


class ButtonStyle(IntEnum):
    """
    Attributes
    ----------
    blurple = 1
    grey    = 2
    green   = 3
    red     = 4
    link    = 5
    """

    primary = 1
    blurple = 1

    secondary = 2
    grey = 2
    gray = 2

    success = 3
    green = 3

    danger = 4
    red = 4

    link = 5


class ButtonPayload(ComponentPayload, total=False):
    style: ButtonStyle
    label: Optional[str]
    emoji: Optional[Union[PartialEmoji, str]]
    url: Optional[str]


class ActionRowPayload(ComponentPayload):
    components: Optional[List[ComponentPayload]]



import discord
from .message_components import *
from .interaction import *


__all__ = ("MessageInteraction", "ButtonInteraction")


class PartialTextChannel:
    def __init__(self, channel_id):
        self.id = channel_id


class MessageInteraction(BaseInteraction):
    """
    Represents a button interaction.
    Obtainable via :class:`discord.Context.wait_for_button_click`
    and in ``on_button_click`` event.

    Attributes
    ----------
    author : :class:`discord.Member` | :class:`discord.User`
        The user that clicked the button
    channel : :class:`discord.Messageable`
        The channel where the click happened
    guild : :class:`discord.Guild` | ``None``
        The guild where the click happened
    message : :class:`discord.Message`
        The message where the button was clicked
    components : :class:`list`
        A list of :class:`ActionRow` instances containing other components
    component : :class:`Component`
        The component that author interacted with
    """

    def __init__(self, client, data):
        super().__init__(client, data)
        state = client._connection

        msg_data = data.get("message")
        if msg_data is None:
            self.message = None
            self.components = []
        else:
            components = msg_data.get("components", [])
            self.components = [ActionRow.from_dict(comp) for comp in components]
            # For some reason "channel_id" might not be included in message data
            if "channel_id" in msg_data:
                channel = client.get_channel(int(msg_data["channel_id"]))
            else:
                msg_data["channel_id"] = self.channel_id
                channel = self.channel
                # For some reason "channel_id" in message reference might not be included
            if "message_reference" in msg_data and "channel_id" not in msg_data["message_reference"]:
                msg_data["message_reference"]["channel_id"] = None if channel is None else channel.id
            # channel must not be None, because channel.id attr is needed in discord.Message.__init__
            self.message = discord.Message(
                state=state, channel=channel or PartialTextChannel(0), data=msg_data  # type: ignore
            )

        component_data = data.get("data", {})
        component_type = component_data.get("component_type", 1)
        custom_id = component_data.get("custom_id")
        self.component: Component = None  # type: ignore
        for action_row in self.components:
            for component in action_row.components:
                if component.custom_id == custom_id and component.type == component_type:
                    self.component = component
                    if component_type == ComponentType.SelectMenu and isinstance(self.component, SelectMenu):
                        self.component._select_options(component_data.get("values", []))
                    break
            if self.component is not None:
                break

    @property
    def clicked_button(self) -> Button:
        if self.component.type == ComponentType.Button and isinstance(self.component, Button):
            return self.component
        return None  # type: ignore

    @property
    def button(self) -> Button:
        return self.clicked_button

    @property
    def select_menu(self) -> SelectMenu:
        if self.component.type == ComponentType.SelectMenu and isinstance(self.component, SelectMenu):
            return self.component
        return None  # type: ignore


ButtonInteraction = MessageInteraction



from __future__ import annotations
import re
import typing
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Callable, Dict, Hashable, List, Optional, TypeVar, Union

import discord

from .app_command_interaction import SlashInteraction
from .types import (
    ApplicationCommandPayload,
    ApplicationCommandPermissionsPayload,
    ApplicationCommandType,
    OptionChoicePayload,
    OptionPayload,
    OptionType,
    RawCommandPermissionPayload,
    SlashCommandPayload,
)

__all__ = (
    "application_command_factory",
    "ApplicationCommandType",
    "ApplicationCommand",
    "SlashCommand",
    "UserCommand",
    "MessageCommand",
    "OptionChoice",
    "Option",
    "OptionParam",
    "option_enum",
    "ApplicationCommandPermissions",
    "SlashCommandPermissions",
    "RawCommandPermission",
    "Type",
)

T_StrFloat = TypeVar('T_StrFloat', str, float)


def application_command_factory(data: ApplicationCommandPayload) -> ApplicationCommand:
    cmd_type = data.get("type", 1)
    if cmd_type == ApplicationCommandType.CHAT_INPUT:
        data = typing.cast(SlashCommandPayload, data)
        return SlashCommand.from_dict(data)
    elif cmd_type == ApplicationCommandType.USER:
        return UserCommand.from_dict(data)
    elif cmd_type == ApplicationCommandType.MESSAGE:
        return MessageCommand.from_dict(data)
    else:
        raise ValueError("Invalid command type")


class OptionChoice:
    """
    Parameters
    ----------
    name : str
        the name of the option-choice (visible to users)
    value : str or int
        the value of the option-choice
    """

    name: str
    value: Any

    def __init__(self, name: str, value: Any) -> None:
        self.name = name
        self.value = value

    def __repr__(self) -> str:
        return "<OptionChoice name='{0.name}' value={0.value}>".format(self)

    def __eq__(self, other):  # type: ignore
        return self.name == other.name and self.value == other.value


class Option:
    """
    Parameters
    ----------
    name : :class:`str`
        option's name
    description : :class:`str`
        option's description
    type : :class:`Type`
        the option type, e.g. ``Type.USER``, see :ref:`option_type`
    required : :class:`bool`
        whether this option is required or not
    choices : List[:class:`OptionChoice`]
        the list of option choices, type :ref:`option_choice`
    options : List[:class:`Option`]
        the list of sub options. You can only specify this parameter if
        the ``type`` is :class:`Type.SUB_COMMAND` or :class:`Type.SUB_COMMAND_GROUP`
    """

    __slots__ = ("name", "description", "type", "required", "choices", "options", "_choice_connectors")

    name: str
    description: Optional[str]
    type: int
    required: bool
    choices: List[OptionChoice]
    options: List[Option]

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        type: int = 3,
        required: bool = False,
        choices: Optional[List[OptionChoice]] = None,
        options: Optional[List[Option]] = None,
    ) -> None:
        assert name.islower(), f"Option name {name!r} must be lowercase"
        self.name = name
        self.description = description
        self.type = type
        self.required = required
        self.choices = choices or []
        if options is not None:
            if self.type == 1:
                for option in options:
                    if option.type < 3:
                        raise ValueError("Unexpected sub_command in a sub_command")
            elif self.type == 2:
                for option in options:
                    if option.type != 1:
                        raise ValueError("Expected sub_command in this sub_command_group")
        self.options = options or []
        self._choice_connectors: Dict[Union[int, str], Any] = {}
        # Wrap choices
        for i, choice in enumerate(self.choices):
            if self.type == Type.INTEGER:
                if not isinstance(choice.value, int):
                    self._choice_connectors[i] = choice.value
                    choice.value = i
            elif self.type == Type.STRING:
                if not isinstance(choice.value, str):
                    valid_value = f"option_choice_{i}"
                    self._choice_connectors[valid_value] = choice.value
                    choice.value = valid_value

    def __repr__(self) -> str:
        string = "name='{0.name}' description='{0.description}' type={0.type} required={0.required}".format(self)
        if self.options:
            string += " options={self.options}"
        if self.choices:
            string += " choices={self.choices}"
        return f"<Option {string}>"

    def __eq__(self, other):  # type: ignore
        return (
            self.name == other.name
            and self.description == other.description
            and self.type == other.type
            and self.required == other.required
            and self.choices == other.choices
            and self.options == other.options
        )

    @classmethod
    def from_dict(cls, payload: OptionPayload) -> Option:
        return Option(
            name=payload["name"],
            description=payload["description"],
            type=payload["type"],
            required=payload.get("required", False),
            options=[
                Option.from_dict(p) for p
                in payload.get("options", ())
            ],
            choices=[
                OptionChoice(**p) for p
                in payload.get("choices", ())
            ],
        )

    def add_choice(self, name: str, value: Any) -> None:
        """
        Adds an OptionChoice to the list of current choices

        Parameters are the same as for :class:`OptionChoice`
        """
        # Wrap the value
        true_value = value
        if self.type == Type.STRING:
            if not isinstance(value, str):
                true_value = f"option_choice_{len(self._choice_connectors)}"
                self._choice_connectors[true_value] = value
        elif self.type == Type.INTEGER:
            if not isinstance(value, int):
                true_value = len(self._choice_connectors)
                self._choice_connectors[true_value] = value
        # Add an option choice
        self.choices.append(OptionChoice(name=name, value=true_value))

    def add_option(
        self,
        name: str,
        description: Optional[str] = None,
        type: int = 3,
        required: bool = False,
        choices: Optional[List[OptionChoice]] = None,
        options: Optional[List[Option]] = None,
    ) -> None:
        """
        Adds an option to the current list of options

        Parameters are the same as for :class:`Option`
        """
        if self.type == 1:
            if type < 3:
                raise ValueError("sub_command can only be nested in a sub_command_group")
        elif self.type == 2:
            if type != 1:
                raise ValueError("Expected sub_command in this sub_command_group")
        self.options.append(
            Option(name=name, description=description, type=type, required=required, choices=choices, options=options)
        )

    def to_dict(self) -> OptionPayload:
        return OptionPayload(
            name=self.name,
            description=self.description,
            type=self.type,
            required=self.required,
            choices=[
                OptionChoicePayload(name=choice.name, value=choice.value)
                for choice in self.choices
            ],
            options=[
                option.to_dict() for
                option in self.options
            ],
        )


class OptionParam:
    """
    Parameters
    ----------
    default : Union[:class:`str`, Callable[[:class:`SlashInteraction`, Any], Any]]
        default value or a default value factory
    name : :class:`str`
        option's name, the parameter name by default
    description : :class:`str`
        option's description
    converter : Callable[[:class:`SlashInteraction`, Any], Any]
        the option's converter, takes in an interaction and the argument
    """

    TYPES: Dict[Hashable, int] = {
        str: 3,
        int: 4,
        bool: 5,
        discord.abc.User: 6,
        discord.User: 6,
        discord.Member: 6,
        discord.abc.GuildChannel: 7,
        discord.TextChannel: 7,
        discord.VoiceChannel: 7,
        discord.CategoryChannel: 7,
        discord.StageChannel: 7,
        discord.StoreChannel: 7,
        discord.Role: 8,
        Union[discord.Member, discord.Role]: 9,
        discord.abc.Snowflake: 9,
        float: 10,
    }

    default: Any = ...
    name: Optional[str]
    description: str
    converter: Optional[Callable[[SlashInteraction, Any], Any]]

    def __init__(
        self,
        default: Any = ...,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
        converter: Optional[Callable[[SlashInteraction, Any], Any]] = None,
    ) -> None:
        self.default = default
        self.name = name
        self.description = description or "-"
        self.converter = converter

    @property
    def required(self) -> bool:
        return self.default is ...

    def __repr__(self) -> str:
        string = "default={0.default} name='{0.name}' description='{0.description}'".format(self)
        return f"<Option {string}>"


def option_param(
    default: Any = ...,
    *,
    name: Optional[str] = None,
    desc: Optional[str] = None,
    description: Optional[str] = None,
    converter: Optional[Callable[[SlashInteraction, Any], Any]] = None,
) -> OptionParam:
    if desc is not None and description is not None:
        raise TypeError("Only desc or description may be used, not both")
    return OptionParam(default, name=name, description=desc or description, converter=converter)


def option_enum(choices: Dict[str, T_StrFloat], **kwargs: T_StrFloat) -> Enum:
    choices = choices or kwargs
    return Enum('', choices, type=type(next(iter(choices.values()))))


class ApplicationCommand(ABC):
    """
    Base class for application commands
    """

    id: int
    name: str
    type: ApplicationCommandType
    application_id: Optional[int] = None

    def __init__(self, type: ApplicationCommandType, **kwargs: Any) -> None:
        self.type = type
        self.id = int(kwargs.get("id", 0))
        application_id = kwargs.get("application_id")
        if application_id is not None:
            self.application_id = int(application_id)
        else:
            self.application_id = application_id

    def __eq__(self, _):  # type: ignore
        return False

    @abstractmethod
    def to_dict(self) -> ApplicationCommandPayload:
        raise NotImplementedError


class UserCommand(ApplicationCommand):
    def __init__(self, name: str, **kwargs: Any) -> None:
        kwargs.pop('type', 0)
        super().__init__(ApplicationCommandType.USER, **kwargs)
        self.name = name

    def __repr__(self) -> str:
        return f"<UserCommand name={self.name!r}>"

    def __eq__(self, other):  # type: ignore
        return self.type == other.type and self.name == other.name

    def to_dict(self) -> ApplicationCommandPayload:
        return {"type": self.type, "name": self.name}

    @classmethod
    def from_dict(cls, data: ApplicationCommandPayload) -> UserCommand:
        if data.get("type", 1) == ApplicationCommandType.USER:
            return UserCommand(**data)
        else:
            raise ValueError(f"{cls.__name__} type can be only {ApplicationCommandType.USER}")


class MessageCommand(ApplicationCommand):
    def __init__(self, name: str, **kwargs: Any) -> None:
        kwargs.pop('type', 0)
        super().__init__(ApplicationCommandType.MESSAGE, **kwargs)
        self.name = name

    def __repr__(self) -> str:
        return f"<MessageCommand name={self.name!r}>"

    def __eq__(self, other):  # type: ignore
        return self.type == other.type and self.name == other.name

    def to_dict(self) -> ApplicationCommandPayload:
        return {"type": self.type, "name": self.name}

    @classmethod
    def from_dict(cls, data: ApplicationCommandPayload) -> MessageCommand:
        if data.get("type", 0) == ApplicationCommandType.MESSAGE:
            return MessageCommand(**data)
        else:
            raise ValueError(f"{cls.__name__} type can be only {ApplicationCommandType.MESSAGE}")


class SlashCommand(ApplicationCommand):
    """
    A base class for building slash-commands.

    Parameters
    ----------
    name : :class:`str`
        The command name
    description : :class:`str`
        The command description (it'll be displayed by discord)
    options : List[Option]
        The options of the command. See :ref:`option`
    default_permission : :class:`bool`
        Whether the command is enabled by default when the app is added to a guild
    """

    options: List[Option]
    default_permission: bool
    permissions: SlashCommandPermissions

    def __init__(
        self,
        name: str,
        description: str,
        options: Optional[List[Option]] = None,
        default_permission: bool = True,
        **kwargs: Any,
    ) -> None:
        kwargs.pop('type', 0)
        super().__init__(ApplicationCommandType.CHAT_INPUT, **kwargs)

        assert (
            re.match(r"^[\w-]{1,32}$", name) is not None and name.islower()
        ), f"Slash command name {name!r} should consist of these symbols: a-z, 0-9, -, _"

        self.name = name
        self.description = description
        self.options = options or []
        self.default_permission = default_permission
        self.permissions = SlashCommandPermissions()

    def __repr__(self) -> str:
        return "<SlashCommand name='{0.name}' description='{0.description}' options={0.options}>".format(self)

    def __eq__(self, other):  # type: ignore
        return (
            self.type == other.type
            and self.name == other.name
            and self.description == other.description
            and self.options == other.options
        )

    @classmethod
    def from_dict(cls, payload: SlashCommandPayload) -> SlashCommand:
        if payload.get("type", 1) != ApplicationCommandType.CHAT_INPUT:
            raise ValueError(f"{cls.__name__} type can be only {ApplicationCommandType.CHAT_INPUT}")

        options = payload.pop("options", [])
        return SlashCommand(
            id=payload["id"],
            name=payload["name"],
            description=payload["description"],
            type=payload["type"],
            default_permission=payload.get("default_permission", True),
            options=[
                Option.from_dict(p) for p in options
            ],
        )

    def add_option(
        self,
        name: str,
        description: Optional[str] = None,
        type: int = 3,
        required: bool = False,
        choices: Optional[List[OptionChoice]] = None,
        options: Optional[List[Option]] = None,
    ) -> None:
        """
        Adds an option to the current list of options

        Parameters are the same as for :class:`Option`
        """
        self.options.append(
            Option(
                name=name,
                description=description,
                type=type,
                required=required,
                choices=choices,
                options=options
            )
        )

    def to_dict(self, *, hide_name: bool = False) -> SlashCommandPayload:
        res: SlashCommandPayload = {
            "type": self.type,
            "description": self.description,
            "options": [o.to_dict() for o in self.options],
        }
        if not self.default_permission:
            res["default_permission"] = False
        if not hide_name:
            res["name"] = self.name
        return res


# Permissions
class ApplicationCommandPermissions:
    """
    Represents application command permissions.
    Roughly equivalent to a list of :class:`RawCommandPermission`

    Application command permissions are checked on the server side.
    Only local application commands can have this type of permissions.

    Parameters
    ----------
    raw_permissions : List[RawCommandPermission]
        a list of :class:`RawCommandPermission`.
        However, :meth:`from_pairs` or :meth:`from_ids`
        might be more convenient.
    """

    permissions: List[RawCommandPermission]

    def __init__(self, raw_permissions: Optional[List[RawCommandPermission]] = None) -> None:
        self.permissions = raw_permissions or []

    def __repr__(self) -> str:
        return "<{0.__name__} permissions={0.permissions!r}>".format(self)

    @classmethod
    def from_pairs(cls, permissions: Dict[Union[discord.Role, discord.User], bool]) -> ApplicationCommandPermissions:
        """
        Creates :class:`ApplicationCommandPermissions` using
        instances of :class:`discord.Role` and :class:`discord.User`

        Parameters
        ----------
        permissions : :class:`dict`
            a dictionary of {:class:`Role | User`: :class:`bool`}
        """
        raw_perms = [RawCommandPermission.from_pair(target, perm) for target, perm in permissions.items()]

        return ApplicationCommandPermissions(raw_perms)

    @classmethod
    def from_ids(
        cls,
        role_perms: Optional[Dict[int, bool]] = None,
        user_perms: Optional[Dict[int, bool]] = None,
    ) -> ApplicationCommandPermissions:
        """
        Creates :class:`ApplicationCommandPermissions` from
        2 dictionaries of IDs and permissions.

        Parameters
        ----------
        role_perms : :class:`dict`
            a dictionary of {``role_id``: :class:`bool`}
        user_perms : :class:`dict`
            a dictionary of {``user_id``: :class:`bool`}
        """
        role_perms = role_perms or {}
        user_perms = user_perms or {}
        raw_perms = [RawCommandPermission(role_id, 1, perm) for role_id, perm in role_perms.items()]

        for user_id, perm in user_perms.items():
            raw_perms.append(RawCommandPermission(user_id, 2, perm))
        return ApplicationCommandPermissions(raw_perms)

    @classmethod
    def from_dict(cls, data: ApplicationCommandPermissionsPayload) -> ApplicationCommandPermissions:
        return SlashCommandPermissions([RawCommandPermission.from_dict(perm) for perm in data["permissions"]])

    def to_dict(self) -> ApplicationCommandPermissionsPayload:
        return {"permissions": [perm.to_dict() for perm in self.permissions]}


class RawCommandPermission:
    """
    Represents command permissions for a role or a user.

    Attributes
    ----------
    id : :class:`int`
        ID of a target
    type : :class:`int`
        1 if target is a role; 2 if target is a user
    permission : :class:`bool`
        allow or deny the access to the command
    """

    __slots__ = ("id", "type", "permission")

    id: int
    type: int
    permission: bool

    def __init__(self, id: int, type: int, permission: bool) -> None:
        self.id = id
        self.type = type
        self.permission = permission

    def __repr__(self) -> str:
        return "<RawCommandPermission id={0.id} type={0.type} permission={0.permission}>".format(self)

    @classmethod
    def from_pair(cls, target: Union[discord.Role, discord.User], permission: bool) -> RawCommandPermission:
        if not isinstance(target, (discord.Role, discord.User)):
            raise discord.InvalidArgument("target should be Role or User")
        if not isinstance(permission, bool):
            raise discord.InvalidArgument("permission should be bool")
        return RawCommandPermission(
            id=target.id, type=1 if isinstance(target, discord.Role) else 2, permission=permission
        )

    @classmethod
    def from_dict(cls, data: RawCommandPermissionPayload) -> RawCommandPermission:
        return RawCommandPermission(**data)

    def to_dict(self) -> RawCommandPermissionPayload:
        return {"id": self.id, "type": self.type, "permission": self.permission}


Type = OptionType
SlashCommandPermissions = ApplicationCommandPermissions



import asyncio
from typing import Any, Awaitable, Callable, Dict, List, Optional, Tuple

import discord
from discord import Guild
from discord.ext import commands
from discord.http import Route

from ..interactions import (
    ApplicationCommand,
    ApplicationCommandPermissions,
    BaseInteraction,
    ComponentType,
    ContextMenuInteraction,
    MessageInteraction,
    SlashCommand,
    SlashInteraction,
    application_command_factory,
)
from ._decohub import _HANDLER
from .context_menus_core import InvokableMessageCommand, InvokableUserCommand, message_command, user_command
from .slash_core import CommandParent, slash_command
from .utils import ClickListener, _on_button_click

__all__ = ("InteractionClient", "SlashClient")


class InteractionClient:
    """
    The main purpose of this class is to track ``INTERACTION_CREATE`` API event.

    Parameters
    ----------
    client : :class:`commands.Bot` | :class:`commands.AutoShardedBot`
        The discord.py Bot instance
    test_guilds : :class:`List[int]`
        A list of IDs of guilds where the application commands
        should be registered instead of global registration
    sync_commands : :class:`bool`
        If set to True, your client will sync all registered commands with the code
    show_warnings : :class:`bool`
        Whether to show the warnings or not. Defaults to ``True``
    modify_send : :class:`bool`
        Whether to modify :class:`Messageable.send` and :class:`Message.edit`.
        Modified methods allow to specify the ``components`` parameter.

    Attributes
    ----------
    client : :class:`commands.Bot` | :class:`commands.AutoShardedBot`
        an instance of any class inherited from :class:`discord.Client`
    application_id : :class:`int`
        the ID of the application your bot is related to
    global_commands : List[:class:`ApplicationCommand`]
        All registered global application commands
    slash_commands : :class:`Dict[str, CommandParent]`
        All invokable slash commands from your code
    user_commands : :class:`Dict[str, InvokableUserCommand]`
        All invokable user commands from your code
    message_commands : :class:`Dict[str, InvokableMessageCommand]`
        All invokable message commands from your code
    commands : :class:`Dict[str, InvokableApplicationCommand]`
        All invokable application commands from your code
    is_ready : bool
        Equals to ``True`` if SlashClient is ready, otherwise it's ``False``
    """

    def __init__(
        self,
        client,
        *,
        test_guilds: List[int] = None,
        sync_commands: bool = True,
        show_warnings: bool = True,
        modify_send: bool = True,
    ) -> None:
        self._uses_discord_2 = hasattr(client, "add_view")
        _HANDLER.client = client
        self.client: Any = _HANDLER.client
        self.application_id = None
        self.events: Dict[str, Callable[..., Awaitable]] = {}
        self._listeners: Dict[Any, List[Tuple[Any, Callable]]] = {}
        self._global_commands: Dict[int, ApplicationCommand] = {}
        self._guild_commands: Dict[int, Dict[int, ApplicationCommand]] = {}
        self._cogs_with_err_listeners = {
            "on_slash_command_error": [],
            "on_user_command_error": [],
            "on_message_command_error": [],
        }
        self._test_guilds = test_guilds
        self._sync_commands = sync_commands
        self._show_warnings = show_warnings
        self._modify_send = modify_send
        self.active_shard_count: int = 0
        self.is_ready: bool = False
        # Add listeners
        self._register_listeners()
        # Modify old discord.py methods
        self._modify_discord()
        # Link the slash ext to client if doesn't exist yet
        if not hasattr(self.client, "slash"):
            self.client.slash = self
        # Inject cogs that are already loaded
        for cog in self.client.cogs.values():
            self._inject_cogs(cog)

    def _register_listeners(self) -> None:
        self.client.add_listener(self._on_guild_remove, "on_guild_remove")
        self.client.add_listener(self._on_socket_response, "on_socket_response")
        if isinstance(self.client, commands.AutoShardedBot):
            self.client.add_listener(self._on_shard_connect, "on_shard_connect")
            self.client.add_listener(self._on_ready, "on_ready")
        else:
            self.client.add_listener(self._on_connect, "on_connect")
        # For nice click listener
        self.client.add_listener(_on_button_click, "on_button_click")

    def _modify_discord(self) -> None:
        # Modify cog loader
        _add_cog = self.client.add_cog

        def add_cog_2(cog):
            self._inject_cogs(cog)
            _add_cog(cog)

        self.client.add_cog = add_cog_2
        # Modify cog unloader
        _rem_cog = self.client.remove_cog

        def rem_cog_2(name):
            self._eject_cogs(name)
            _rem_cog(name)

        self.client.remove_cog = rem_cog_2
        # Multiple wait for
        self.client.multiple_wait_for = self.multiple_wait_for
        # Change other class methods

        async def ctx_wait_for_button_click(ctx, check=None, timeout=None):
            return await self.wait_for_button_click(check=check, timeout=timeout)

        async def message_wait_for_button_click(message, check=None, timeout=None):
            if check is None:
                check = lambda inter: True

            def auto_check(inter):
                if message.id != inter.message.id:
                    return False
                return check(inter)

            return await self.wait_for_button_click(auto_check, timeout)

        async def message_wait_for_dropdown(message, check=None, timeout=None):
            if check is None:
                check = lambda inter: True

            def auto_check(inter):
                if message.id != inter.message.id:
                    return False
                return check(inter)

            return await self.wait_for_dropdown(auto_check, timeout)

        async def fetch_commands(guild):
            return await self.fetch_guild_commands(guild.id)

        async def fetch_command(guild, command_id):
            return await self.fetch_guild_command(guild.id, command_id)

        async def edit_command(guild, command_id, slash_command):
            return await self.edit_guild_slash_command(guild.id, command_id, slash_command)  # type: ignore

        async def edit_command_permissions(guild, command_id, permissions):
            return await self.edit_guild_command_permissions(guild.id, command_id, permissions)

        async def batch_edit_command_permissions(guild, permissions):
            return await self.batch_edit_guild_command_permissions(guild.id, permissions)

        async def delete_command(guild, command_id):
            return await self.delete_guild_command(guild.id, command_id)

        async def delete_commands(guild):
            return await self.delete_guild_commands(guild.id)

        def get_commands(guild):
            return self.get_guild_commands(guild.id)

        def get_command(guild, command_id):
            return self.get_guild_command(guild.id, command_id)

        def get_command_named(guild, name):
            return self.get_guild_command_named(guild.id, name)

        def create_click_listener(message, timeout=None):
            return ClickListener(message.id, timeout)

        # hack to allow monkey patching by declaring a module as Any
        discord = globals()["discord"]
        commands = globals()["commands"]

        if self._modify_send:
            if self._uses_discord_2:
                from ._modifications.new import edit as edit_with_components
                from ._modifications.new import send as send_with_components
            else:
                from ._modifications.old import (
                    create_message_with_components,
                    edit_with_components,
                    send_with_components,
                )

                discord.state.create_message = create_message_with_components
            discord.abc.Messageable.send = send_with_components
            discord.Message.edit = edit_with_components

        commands.Context.wait_for_button_click = ctx_wait_for_button_click
        discord.Message.create_click_listener = create_click_listener
        discord.Message.wait_for_button_click = message_wait_for_button_click
        discord.Message.wait_for_dropdown = message_wait_for_dropdown
        discord.Guild.get_commands = get_commands
        discord.Guild.get_command = get_command
        discord.Guild.get_command_named = get_command_named
        discord.Guild.fetch_commands = fetch_commands
        discord.Guild.fetch_command = fetch_command
        discord.Guild.edit_command = edit_command
        discord.Guild.edit_command_permissions = edit_command_permissions
        discord.Guild.batch_edit_command_permissions = batch_edit_command_permissions
        discord.Guild.delete_command = delete_command
        discord.Guild.delete_commands = delete_commands

    def teardown(self) -> None:
        """Cleanup the client by removing all registered listeners and caches."""
        self.client.remove_listener(self._on_guild_remove, "on_guild_remove")
        self.client.remove_listener(self._on_socket_response, "on_socket_response")
        if isinstance(self.client, commands.AutoShardedBot):
            self.client.remove_listener(self._on_shard_connect, "on_shard_connect")
            self.client.remove_listener(self._on_ready, "on_ready")
        else:
            self.client.remove_listener(self._on_connect, "on_connect")

        self.events.clear()
        self._listeners.clear()
        self._global_commands.clear()
        self._guild_commands.clear()
        if hasattr(self.client, "slash"):
            del self.client.slash  # type: ignore
        self.is_ready = False

    @property
    def slash_commands(self) -> Dict[str, CommandParent]:
        return _HANDLER.slash_commands

    @property
    def user_commands(self) -> Dict[str, CommandParent]:
        return _HANDLER.user_commands

    @property
    def message_commands(self) -> Dict[str, CommandParent]:
        return _HANDLER.message_commands

    @property
    def commands(self) -> Dict[str, CommandParent]:
        return dict(**_HANDLER.slash_commands, **_HANDLER.user_commands, **_HANDLER.message_commands)

    @property
    def global_commands(self) -> List[ApplicationCommand]:
        return [sc for sc in self._global_commands.values()]

    def event(self, func: Callable[..., Awaitable]) -> Callable[..., Awaitable]:
        """
        Decorator
        ::

            @slash.event
            async def on_ready():
                print("SlashClient is ready")

        | All possible events:
        | ``on_ready``, ``on_auto_register``,
        | ``on_slash_command``, ``on_slash_command_error``
        """
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f"<{func.__qualname__}> must be a coroutine function")
        name = func.__name__
        if name.startswith("on_"):
            name = name[3:]
            self.events[name] = func
        return func

    def slash_command(self, *args, **kwargs) -> Callable[[Callable[..., Awaitable]], CommandParent]:
        """
        A decorator that allows to build a slash command.

        Parameters
        ----------
        auto_sync : :class:`bool`
            whether to automatically register the command or not. Defaults to ``True``
        name : :class:`str`
            name of the slash command you want to respond to (equals to function name by default).
        description : :class:`str`
            the description of the slash command. It will be visible in Discord.
        options : :class:`List[Option]`
            the list of slash command options. The options will be visible in Discord.
        default_permission : :class:`bool`
            whether the command is enabled by default when the app is added to a guild.
        guild_ids : :class:`List[int]`
            if specified, the client will register a command in these guilds.
            Otherwise this command will be registered globally.
        connectors : :class:`dict`
            which function param states for each option. If the name
            of an option already matches the corresponding function param,
            you don't have to specify the connectors. Connectors template:
            ``{"option-name": "param_name", ...}``
        """
        return slash_command(*args, **kwargs)

    def user_command(self, *args, **kwargs) -> Callable[[Callable[..., Awaitable]], InvokableUserCommand]:
        return user_command(*args, **kwargs)

    def message_command(self, *args, **kwargs) -> Callable[[Callable[..., Awaitable]], InvokableMessageCommand]:
        return message_command(*args, **kwargs)

    # Getters
    def get_global_command(self, command_id: int):
        """
        Get a cached global command

        Parameters
        ----------

        command_id : int
            the ID of the command

        Returns
        -------

        slash_command : SlashCommand | None
        """
        return self._global_commands.get(command_id)

    def get_global_command_named(self, name: str):
        """
        Get a cached global command matching the specified name

        Parameters
        ----------

        name : str
            the name of the command

        Returns
        -------

        slash_command : SlashCommand | None
        """
        for cmd in self._global_commands.values():
            if cmd and cmd.name == name:
                return cmd

    def get_guild_command(self, guild_id: int, command_id: int):
        """
        Get a cached guild command

        Parameters
        ----------

        guild_id : int
            the ID of the guild

        command_id : int
            the ID of the command

        Returns
        -------

        slash_command : SlashCommand | None
        """
        granula = self._guild_commands.get(guild_id)
        if granula is not None:
            return granula.get(command_id)

    def get_guild_command_named(self, guild_id: int, name: str):
        """
        Get a cached guild command matching the specified name

        Parameters
        ----------

        guild_id : int
            the ID of the guild

        name : str
            the name of the command

        Returns
        -------

        slash_command : SlashCommand | None
        """
        granula = self._guild_commands.get(guild_id)
        if granula is not None:
            for cmd in granula.values():
                if cmd.name == name:
                    return cmd

    def get_guild_commands(self, guild_id: int):
        """
        Get cached guild commands

        Parameters
        ----------
        guild_id : int
            the ID of the guild

        Returns
        -------
        ~:class:`List[ApplicationCommand]`
        """
        granula = self._guild_commands.get(guild_id, {})
        return [sc for sc in granula.values()]

    # Straight references to API
    async def fetch_global_commands(self):
        """
        Requests a list of global registered commands from the API

        Returns
        -------

        global_commands : List[ApplicationCommand]
        """
        data = await self.client.http.request(
            Route("GET", "/applications/{application_id}/commands", application_id=self.application_id)
        )
        return [application_command_factory(dat) for dat in data]

    async def fetch_guild_commands(self, guild_id: int):
        """
        Requests a list of registered commands for a specific guild

        Parameters
        ----------

        guild_id : int

        Returns
        -------

        guild_commands : List[ApplicationCommand]
        """
        data = await self.client.http.request(
            Route(
                "GET",
                "/applications/{application_id}/guilds/{guild_id}/commands",
                application_id=self.application_id,
                guild_id=guild_id,
            )
        )
        return [application_command_factory(dat) for dat in data]

    async def fetch_global_command(self, command_id: int):
        """
        Requests a registered global command

        Parameters
        ----------

        command_id : int

        Returns
        -------

        global_command : ApplicationCommand
        """
        data = await self.client.http.request(
            Route(
                "GET",
                "/applications/{application_id}/commands/{cmd_id}",
                application_id=self.application_id,
                cmd_id=command_id,
            )
        )
        return application_command_factory(data)

    async def fetch_guild_command(self, guild_id: int, command_id: int):
        """
        Requests a registered guild command

        Parameters
        ----------

        guild_id : int

        command_id : int

        Returns
        -------

        guild_command : ApplicationCommand
        """
        data = await self.client.http.request(
            Route(
                "GET",
                "/applications/{application_id}/guilds/{guild_id}/commands/{cmd_id}",
                application_id=self.application_id,
                guild_id=guild_id,
                cmd_id=command_id,
            )
        )
        return application_command_factory(data)

    async def register_global_command(self, app_command: ApplicationCommand):
        """
        Registers a global application command

        Parameters
        ----------

        app_command : ApplicationCommand
        """
        if not isinstance(app_command, ApplicationCommand):
            raise discord.InvalidArgument("Expected an ApplicationCommand instance")
        r = await self.client.http.request(
            Route("POST", "/applications/{application_id}/commands", application_id=self.application_id),
            json=app_command.to_dict(),
        )
        sc = application_command_factory(r)
        self._global_commands[sc.id] = sc
        return sc

    async def register_guild_command(self, guild_id: int, app_command: ApplicationCommand):
        """
        Registers a local application command

        Parameters
        ----------

        guild_id : :class:`int`

        app_command : :class:`ApplicationCommand`
        """
        if not isinstance(app_command, ApplicationCommand):
            raise discord.InvalidArgument("Expected a ApplicationCommand instance")
        r = await self.client.http.request(
            Route(
                "POST",
                "/applications/{app_id}/guilds/{guild_id}/commands",
                app_id=self.application_id,
                guild_id=guild_id,
            ),
            json=app_command.to_dict(),
        )
        # Update cache
        sc = application_command_factory(r)
        self._add_guild_command(guild_id, sc)
        return sc

    async def overwrite_global_commands(self, app_commands: list):
        """
        Bulk overwrites all global application commands

        Parameters
        ----------

        app_commands : List[ApplicationCommand]
        """

        if not all(isinstance(sc, ApplicationCommand) for sc in app_commands):
            raise discord.InvalidArgument("app_commands must contain only ApplicationCommand instances")
        await self.client.http.request(
            Route("PUT", "/applications/{app_id}/commands", app_id=self.application_id),
            json=[sc.to_dict() for sc in app_commands],
        )
        # Update cache
        new_commands = await self.fetch_global_commands()
        self._global_commands = {cmd.id: cmd for cmd in new_commands}
        return new_commands

    async def overwrite_guild_commands(self, guild_id: int, app_commands: List[ApplicationCommand]):
        """
        Bulk overwrites all guild application commands

        Parameters
        ----------

        guild_id : int

        app_commands : List[ApplicationCommand]
        """
        if not all(isinstance(sc, ApplicationCommand) for sc in app_commands):
            raise discord.InvalidArgument("app_commands must contain only ApplicationCommand instances")
        await self.client.http.request(
            Route(
                "PUT",
                "/applications/{app_id}/guilds/{guild_id}/commands",
                app_id=self.application_id,
                guild_id=guild_id,
            ),
            json=[sc.to_dict() for sc in app_commands],
        )
        # Update cache
        new_commands = await self.fetch_guild_commands(guild_id)
        self._guild_commands[guild_id] = {cmd.id: cmd for cmd in new_commands}

    async def edit_global_command(self, command_id: int, app_command: ApplicationCommand, **kwargs):
        """
        Edits a global application command

        Parameters
        ----------
        command_id : int
        app_command : ApplicationCommand
            replacement of the old data
        """
        if not isinstance(app_command, ApplicationCommand):
            raise discord.InvalidArgument("parameter app_command must be ApplicationCommand")
        ignore_name = kwargs.get("ignore_name", False)
        r = await self.client.http.request(
            Route("PATCH", "/applications/{app_id}/commands/{cmd_id}", app_id=self.application_id, cmd_id=command_id),
            json=app_command.to_dict(hide_name=ignore_name),
        )
        # Update cache
        sc = application_command_factory(r)
        self._global_commands[sc.id] = sc
        return sc

    async def edit_guild_command(self, guild_id: int, command_id: int, app_command: ApplicationCommand, **kwargs):
        """
        Edits the local application command

        Parameters
        ----------
        guild_id : int
        command_id : int
        app_command : ApplicationCommand
            replacement of the old data
        """
        if not isinstance(app_command, ApplicationCommand):
            raise discord.InvalidArgument("parameter app_command must be ApplicationCommand")
        ignore_name = kwargs.get("ignore_name", False)
        r = await self.client.http.request(
            Route(
                "PATCH",
                "/applications/{app_id}/guilds/{guild_id}/commands/{cmd_id}",
                app_id=self.application_id,
                guild_id=guild_id,
                cmd_id=command_id,
            ),
            json=app_command.to_dict(hide_name=ignore_name),
        )
        # Update cache
        sc = application_command_factory(r)
        self._add_guild_command(guild_id, sc)
        return sc

    async def delete_global_command(self, command_id: int):
        """
        Deletes the global application command

        Parameters
        ----------
        command_id : int
        """
        await self.client.http.request(
            Route("DELETE", "/applications/{app_id}/commands/{cmd_id}", app_id=self.application_id, cmd_id=command_id)
        )
        # Update cache
        self._remove_global_command(command_id)

    async def delete_guild_command(self, guild_id: int, command_id: int):
        """
        Deletes the local application command

        Parameters
        ----------
        guild_id : int
        command_id : int
        """
        await self.client.http.request(
            Route(
                "DELETE",
                "/applications/{app_id}/guilds/{guild_id}/commands/{cmd_id}",
                app_id=self.application_id,
                guild_id=guild_id,
                cmd_id=command_id,
            )
        )
        # Update cache
        self._remove_guild_command(guild_id, command_id)

    async def delete_global_commands(self):
        """
        Deletes all global commands.
        """
        await self.overwrite_global_commands([])
        self._global_commands = {}

    async def delete_guild_commands(self, guild_id: int):
        """
        Deletes all local commands in the specified guild.

        Parameters
        ----------
        guild_id : int
            the ID of the guild where you're going to delete the commands
        """
        await self.overwrite_guild_commands(guild_id, [])
        if guild_id in self._guild_commands:
            del self._guild_commands[guild_id]

    # Permissions
    async def fetch_guild_command_permissions(self, guild_id: int, command_id: int):
        """
        Fetches command permissions for a specific command in a guild.

        Parameters
        ----------
        guild_id : :class:`int`
            the ID of the guild
        command_id : :class:`int`
            the ID of the command
        """
        r = await self.client.http.request(
            Route(
                "GET",
                "/applications/{app_id}/guilds/{guild_id}/commands/{command_id}/permissions",
                app_id=self.application_id,
                guild_id=guild_id,
                command_id=command_id,
            )
        )
        return ApplicationCommandPermissions.from_dict(r)

    async def batch_fetch_guild_command_permissions(self, guild_id: int):
        """
        Fetches command permissions for all commands in a guild.

        Parameters
        ----------
        guild_id : :class:`int`
            the ID of the guild
        """
        array = await self.client.http.request(
            Route(
                "GET",
                "/applications/{app_id}/guilds/{guild_id}/commands/permissions",
                app_id=self.application_id,
                guild_id=guild_id,
            )
        )
        return {int(obj["id"]): ApplicationCommandPermissions.from_dict(obj) for obj in array}

    async def edit_guild_command_permissions(
        self, guild_id: int, command_id: int, permissions: ApplicationCommandPermissions
    ):
        """
        Edits command permissions for a specific command in a guild.

        Parameters
        ----------
        guild_id : :class:`int`
            the ID of the guild
        command_id : :class:`int`
            the ID of the command
        permissions : :class:`SlashCommandPermissions` | :class:`dict`
            new permissions to set. If you use :class:`SlashCommandPermissions.from_pairs`,
            you can pass the arg of that method straight into this function
        """

        if isinstance(permissions, dict):
            permissions = ApplicationCommandPermissions.from_pairs(permissions)

        await self.client.http.request(
            Route(
                "PUT",
                "/applications/{app_id}/guilds/{guild_id}/commands/{command_id}/permissions",
                app_id=self.application_id,
                guild_id=guild_id,
                command_id=command_id,
            ),
            json=permissions.to_dict(),
        )
        # Update cache
        self._set_permissions(guild_id, command_id, permissions)

    async def batch_edit_guild_command_permissions(
        self, guild_id: int, permissions: Dict[int, ApplicationCommandPermissions]
    ):
        """
        Batch edits permissions for all commands in a guild.

        Parameters
        ----------
        guild_id : :class:`int`
            the ID of the guild
        permissions : Dict[:class:`int`, :class:`ApplicationCommandPermissions`]
            a dictionary of command IDs and permissions
        """
        data = []
        for cmd_id, perms in permissions.items():
            # Cache
            self._set_permissions(guild_id, cmd_id, perms)
            # To dict
            thing = {**perms.to_dict(), "id": cmd_id}
            data.append(thing)
        await self.client.http.request(
            Route(
                "PUT",
                "/applications/{app_id}/guilds/{guild_id}/commands/permissions",
                app_id=self.application_id,
                guild_id=guild_id,
            ),
            json=data,
        )

    # Even slower API methods
    async def fetch_global_command_named(self, name: str) -> Optional[ApplicationCommand]:
        """
        Fetches a global command that matches the specified name

        Parameters
        ----------

        name : str
            the name of the command to fetch
        """
        cmds = await self.fetch_global_commands()
        for c in cmds:
            if c.name == name:
                return c

    async def fetch_guild_command_named(self, guild_id: int, name: str):
        """
        Fetches a guild command that matches the specified name

        Parameters
        ----------

        guild_id : int
            ID of the guild where the command is registered

        name : str
            the name of the command to fetch
        """
        cmds = await self.fetch_guild_commands(guild_id)
        for cmd in cmds:
            if cmd.name == name:
                return cmd

    async def edit_global_command_named(self, name: str, app_command: ApplicationCommand):
        """
        Edits a global command matching the specified name.

        Parameters
        ----------
        name : str
            the name of the command to edit
        app_command : ApplicationCommand
            replacement of the old data
        """
        cmd = self.get_global_command_named(name)
        if cmd is not None:
            await self.edit_global_command(cmd.id, app_command)

    async def edit_guild_command_named(self, guild_id: int, name: str, app_command: ApplicationCommand):
        """
        Edits a local command matching the specified name.

        Parameters
        ----------
        guild_id : int
            ID of the guild where the command is registered
        name : str
            the name of the command to edit
        app_command : ApplicationCommand
            replacement of the old data
        """
        cmd = self.get_guild_command_named(guild_id, name)
        if cmd is not None:
            await self.edit_guild_command(guild_id, cmd.id, app_command)

    async def delete_global_command_named(self, name: str):
        """
        Deletes a global command matching the specified name.

        Parameters
        ----------
        name : str
            the name of the command to delete
        """
        cmd = self.get_global_command_named(name)
        if cmd is not None:
            await self.delete_global_command(cmd.id)

    async def delete_guild_command_named(self, guild_id: int, name: str):
        """
        Deletes a local command matching the specified name.

        Parameters
        ----------
        guild_id : int
            ID of the guild where the command is registered
        name : str
            the name of the command to edit
        """
        cmd = self.get_guild_command_named(guild_id, name)
        if cmd is not None:
            await self.delete_guild_command(guild_id, cmd.id)

    # Internal things
    def dispatch(self, event_name: str, *args, **kwargs):
        self.client.loop.create_task(self._activate_event(event_name, *args, **kwargs))

    def _add_global_command(self, command: ApplicationCommand):
        self._global_commands[command.id] = command

    def _add_guild_command(self, guild_id: int, command: ApplicationCommand):
        if guild_id not in self._guild_commands:
            self._guild_commands[guild_id] = {command.id: command}
        else:
            self._guild_commands[guild_id][command.id] = command

    def _remove_global_command(self, command_id: int):
        if command_id in self._global_commands:
            del self._global_commands[command_id]

    def _remove_guild_command(self, guild_id: int, command_id: int):
        if guild_id in self._guild_commands:
            granula = self._guild_commands[guild_id]
            if command_id in granula:
                del granula[command_id]

    def _set_permissions(self, guild_id: int, command_id: int, permissions: ApplicationCommandPermissions):
        if guild_id in self._guild_commands:
            granula = self._guild_commands[guild_id]
            if command_id in granula:
                command = granula[command_id]
                if isinstance(command, SlashCommand):
                    command.permissions = permissions

    def _error_handler_exists(self, handler_name: str) -> bool:
        cog_names = self._cogs_with_err_listeners.get(handler_name, [])
        return not (
            len(cog_names) == 0
            and not hasattr(self.client, handler_name)
            and handler_name not in self.client._listeners
            and handler_name not in self._listeners
            and handler_name not in self.events
        )

    def _inject_cogs(self, cog: Any):
        # Insert the cog into slash commands:
        if not hasattr(cog, "slash_commands"):
            cog.slash_commands = []
        for cmd in self.slash_commands.values():
            if cmd._cog_class_name == cog.__class__.__name__:
                cmd._inject_cog(cog)
                cog.slash_commands.append(cmd)
        # Insert the cog into user commands:
        if not hasattr(cog, "user_commands"):
            cog.user_commands = []
        for cmd in self.user_commands.values():
            if cmd._cog_class_name == cog.__class__.__name__:
                cmd._inject_cog(cog)
                cog.user_commands.append(cmd)
        # Insert the cog into message commands:
        if not hasattr(cog, "message_commands"):
            cog.message_commands = []
        for cmd in self.message_commands.values():
            if cmd._cog_class_name == cog.__class__.__name__:
                cmd._inject_cog(cog)
                cog.message_commands.append(cmd)
        # Auto register the commands again
        if self.is_ready:
            self.client.loop.create_task(self._auto_register_or_patch())
        # We need to know which cogs are able to dispatch errors
        pairs = cog.get_listeners()
        for event_name, func in pairs:
            cog_names = self._cogs_with_err_listeners.get(event_name)
            if cog_names is not None:
                cog_names.append(cog.qualified_name)

    def _eject_cogs(self, name: str):
        for cog_names in self._cogs_with_err_listeners.values():
            try:
                cog_names.remove(name)
            except Exception:
                pass
        # Remove the commands from cache
        bad_keys = [kw for kw, cmd in _HANDLER.slash_commands.items() if cmd._cog_name == name]
        for key in bad_keys:
            del _HANDLER.slash_commands[key]
        bad_keys = [kw for kw, cmd in _HANDLER.user_commands.items() if cmd._cog_name == name]
        for key in bad_keys:
            del _HANDLER.user_commands[key]
        bad_keys = [kw for kw, cmd in _HANDLER.message_commands.items() if cmd._cog_name == name]
        for key in bad_keys:
            del _HANDLER.message_commands[key]

    def _guilds_with_commands(self) -> List[int]:
        guilds = set()
        for cmd in _HANDLER.slash_commands.values():
            guild_ids = cmd.guild_ids or self._test_guilds
            if guild_ids is not None:
                guilds = guilds.union(set(guild_ids))
        return list(guilds)

    def _per_guild_commands(self) -> Tuple[List[SlashCommand], Dict[int, List[SlashCommand]]]:
        global_cmds = []
        guilds = {}
        for cmd in self.commands.values():
            if not cmd.auto_sync:
                continue
            guild_ids = cmd.guild_ids or self._test_guilds
            if guild_ids is None:
                global_cmds.append(cmd.registerable)
            else:
                for guild_id in guild_ids:
                    if guild_id not in guilds:
                        guilds[guild_id] = [cmd.registerable]
                    else:
                        guilds[guild_id].append(cmd.registerable)
        return global_cmds, guilds

    def _modify_parser(self, parsers: Dict[str, Callable[..., Any]], event: str, func: Callable[[Any], Any]):
        def empty_func(data):
            pass

        old_func: Callable = parsers.get(event, empty_func)
        original_func = getattr(old_func, "__original_parser__", old_func)

        def new_func(data):
            func(data)
            return original_func(data)

        new_func.__original_parser__ = original_func
        parsers[event] = new_func

    # Automatically register commands
    async def _auto_register_or_patch(self):
        """
        Assuming that all commands are already cached
        ---------------------------------------------
        """
        if not self._sync_commands:
            return
        # Sort all invokable commands between guild IDs
        global_cmds, guild_cmds = self._per_guild_commands()
        # This is for the event
        global_commands_patched = False
        patched_guilds = []
        # Update global commands first
        update_required = False
        deletion_required = False
        for cmd in global_cmds:
            old_cmd = self.get_global_command_named(cmd.name)
            if old_cmd is None or old_cmd.type == cmd.type and cmd != old_cmd:
                update_required = True
                break
            elif old_cmd.type != cmd.type:
                update_required = True
                deletion_required = True
                break
        if update_required or len(global_cmds) != len(self._global_commands):
            try:
                if deletion_required:
                    await self.delete_global_commands()
                await self.overwrite_global_commands(global_cmds)
                global_commands_patched = True
            except Exception as e:
                if self._show_warnings:
                    print(f"[WARNING] Failed to overwrite global commands due to {e}")
        # Update guild commands
        for guild_id, cmds in guild_cmds.items():
            update_required = False
            deletion_required = False
            for cmd in cmds:
                old_cmd = self.get_guild_command_named(guild_id, cmd.name)
                if old_cmd is None or old_cmd.type == cmd.type and cmd != old_cmd:
                    update_required = True
                    break
                elif old_cmd.type != cmd.type:
                    update_required = True
                    deletion_required = True
                    break
            if update_required or len(cmds) != len(self.get_guild_commands(guild_id)):
                try:
                    if deletion_required:
                        await self.delete_guild_commands(guild_id)
                    await self.overwrite_guild_commands(guild_id, cmds)  # type: ignore
                    patched_guilds.append(guild_id)
                except Exception as e:
                    if self._show_warnings:
                        print(f"[WARNING] Failed to overwrite commands in <Guild id={guild_id}> due to {e}")
        # Dispatch an event
        self.dispatch("auto_register", global_commands_patched, patched_guilds)

    async def _maybe_unregister_commands(self, guild_id: Optional[int]):
        """
        Unregisters unknown commands from the guild, if possible.
        Mainly called if a guild command isn't in the code, but
        it still exists in discord and creates interactions.
        """
        if guild_id is None:
            return

        app_commands = await self.fetch_guild_commands(guild_id)
        local_app_commands = self.get_guild_commands(guild_id)
        good_commands = [cmd for cmd in app_commands if cmd.name in local_app_commands]
        try:
            await self.overwrite_guild_commands(guild_id, good_commands)
        except Exception:
            pass

    # Cache commands
    async def _cache_global_commands(self):
        commands = await self.fetch_global_commands()
        self._global_commands = {cmd.id: cmd for cmd in commands}

    async def _cache_guild_commands(self):
        for guild_id in self._guilds_with_commands():
            try:
                commands = await self.fetch_guild_commands(guild_id)
                perms = await self.batch_fetch_guild_command_permissions(guild_id)
                if len(commands) > 0:
                    # Merge commands and permissions
                    merged_commands = {}
                    for cmd in commands:
                        if isinstance(cmd, SlashCommand) and cmd.id in perms:
                            cmd.permissions = perms[cmd.id]
                        merged_commands[cmd.id] = cmd
                    # Put to the dict
                    self._guild_commands[guild_id] = merged_commands
            except Exception:
                pass

    # Special waiter
    async def wait_for_button_click(self, check: Callable[..., bool] = None, timeout: float = None):
        if check is None:
            check = lambda ctx: True
        future = self.client.loop.create_future()
        if "button_click" not in self._listeners:
            self._listeners["button_click"] = []
        listeners = self._listeners["button_click"]
        listeners.append((future, check))
        return await asyncio.wait_for(future, timeout)

    async def wait_for_dropdown(self, check: Callable[..., bool] = None, timeout: float = None):
        if check is None:
            check = lambda ctx: True
        future = self.client.loop.create_future()
        if "dropdown" not in self._listeners:
            self._listeners["dropdown"] = []
        listeners = self._listeners["dropdown"]
        listeners.append((future, check))
        return await asyncio.wait_for(future, timeout)

    async def multiple_wait_for(self, events_and_checks: Dict[str, Any], timeout: float = None):
        """
        Waits until one of the given events toggles and matches the relevant check.

        Example:

        ::

            result = None
            try:
                result = await client.multiple_wait_for(
                    {
                        "message": lambda msg: msg.author == ctx.author,
                        "reaction_add": lambda react, user: user == ctx.author
                    },
                    timeout=60
                )
            except asyncio.TimeoutError:
                await ctx.send("It took too long")
            if isinstance(result, discord.Message):
                # on_message was toggled
                await ctx.send(f"You said '{result.content}'")
            else:
                # on_reaction_add was toggled
                reaction, user = result
                await ctx.send(f"Your reaction: {reaction.emoji}")

        Parameters
        ----------
        events_and_checks : Dict[:class:`str`, :class:`function | None`]
            a dictionary of event names and relevant checks, e.g.
            ``{"message": lambda m: m.author == ctx.author, "button_click": None}``
        timeout : :class:`float` | :class:`None`
            the amount of seconds the bot should wait for any of the given events
        """

        coros_and_events = {}
        waiters = []
        for event, check in events_and_checks.items():
            coro = self.client.wait_for(event, check=check, timeout=timeout)
            coros_and_events[id(coro)] = event
            waiters.append(coro)
        # Wait for some of the waiters to toggle
        done, pending = await asyncio.wait(waiters, return_when=asyncio.FIRST_COMPLETED)
        # Get the result or catch an exception
        try:
            stuff = done.pop().result()
            err = None
        except Exception as error:
            stuff = None
            err = error
        # Clean other waiters
        for future in done:
            future.exception()
        for future in pending:
            future.cancel()
        # Error or result
        if err is not None:
            raise err
        return stuff

    # Adding relevant listeners
    def _on_raw_interaction(self, data: Dict[str, Any]):
        self.client.loop.create_task(self._process_interaction(data))

    async def _on_socket_response(self, payload: Dict[str, Any]):
        if payload.get("t") != "INTERACTION_CREATE":
            return
        await self._process_interaction(payload["d"])

    async def _on_shard_connect(self, shard_id: int):
        self.active_shard_count += 1
        if self.active_shard_count == 1:
            await self._fill_app_id()
            await self._cache_global_commands()
            await self._cache_guild_commands()
            await self._auto_register_or_patch()
        if self._uses_discord_2:
            self._modify_parser(
                self.client._AutoShardedClient__shards[shard_id].ws._discord_parsers,
                "INTERACTION_CREATE",
                self._on_raw_interaction,
            )

    async def _on_connect(self):
        if not isinstance(self.client, discord.AutoShardedClient):
            if self._uses_discord_2:
                self._modify_parser(self.client.ws._discord_parsers, "INTERACTION_CREATE", self._on_raw_interaction)
            await self._fill_app_id()
            await self._cache_global_commands()
            await self._cache_guild_commands()
            await self._auto_register_or_patch()
            self.is_ready = True
            await self._activate_event("ready")

    async def _on_ready(self):
        if isinstance(self.client, discord.AutoShardedClient):
            self.is_ready = True
            await self._activate_event("ready")

    async def _on_guild_remove(self, guild: Guild):
        if guild.id in self._guild_commands:
            del self._guild_commands[guild.id]

    async def _on_slash_command(self, inter: SlashInteraction):
        app_command = self.slash_commands.get(inter.data.name)
        if app_command is None:
            await self._maybe_unregister_commands(inter.guild_id)
            return
        else:
            guild_ids = app_command.guild_ids or self._test_guilds
            is_global = self.get_global_command(inter.data.id) is not None
            if guild_ids is None:
                usable = is_global
            else:
                usable = not is_global and inter.guild_id in guild_ids
        if usable:
            try:
                await app_command.invoke(inter)
            except Exception as err:
                if self._error_handler_exists("on_slash_command_error"):
                    await self._activate_event("slash_command_error", inter, err)
                elif app_command._error_handler is None:
                    raise err
        else:
            await self._maybe_unregister_commands(inter.guild_id)

    async def _on_user_command(self, inter: ContextMenuInteraction):
        app_command = _HANDLER.user_commands.get(inter.data.name)
        if app_command is None:
            await self._maybe_unregister_commands(inter.guild_id)
            return
        else:
            guild_ids = app_command.guild_ids or self._test_guilds
            is_global = self.get_global_command(inter.data.id) is not None
            if guild_ids is None:
                usable = is_global
            else:
                usable = not is_global and inter.guild_id in guild_ids
        if usable:
            try:
                await app_command.invoke(inter)
            except Exception as err:
                if self._error_handler_exists("on_user_command_error"):
                    await self._activate_event("user_command_error", inter, err)
                elif app_command._error_handler is None:
                    raise err
        else:
            await self._maybe_unregister_commands(inter.guild_id)

    async def _on_message_command(self, inter: ContextMenuInteraction):
        app_command = _HANDLER.message_commands.get(inter.data.name)
        if app_command is None:
            await self._maybe_unregister_commands(inter.guild_id)
            return
        else:
            guild_ids = app_command.guild_ids or self._test_guilds
            is_global = self.get_global_command(inter.data.id) is not None
            if guild_ids is None:
                usable = is_global
            else:
                usable = not is_global and inter.guild_id in guild_ids
        if usable:
            try:
                await app_command.invoke(inter)
            except Exception as err:
                if self._error_handler_exists("on_message_command_error"):
                    await self._activate_event("message_command_error", inter, err)
                elif app_command._error_handler is None:
                    raise err
        else:
            await self._maybe_unregister_commands(inter.guild_id)

    async def _toggle_listeners(self, event: str, *args, **kwargs):
        listeners = self._listeners.get(event)
        if listeners:
            removed = []
            for i, (future, condition) in enumerate(listeners):
                if future.cancelled():
                    removed.append(i)
                    continue

                try:
                    result = condition(*args)
                except Exception as exc:
                    future.set_exception(exc)
                    removed.append(i)
                else:
                    if result:
                        if len(args) == 0:
                            future.set_result(None)
                        elif len(args) == 1:
                            future.set_result(args[0])
                        else:
                            future.set_result(args)
                        removed.append(i)
            if len(removed) == len(listeners):
                self._listeners.pop(event)
            else:
                for idx in reversed(removed):
                    del listeners[idx]

    async def _activate_event(self, event_name: str, *args, **kwargs):
        await self._toggle_listeners(event_name, *args, **kwargs)
        if event_name != "ready":
            self.client.dispatch(event_name, *args, **kwargs)
        func = self.events.get(event_name)
        if func:
            await func(*args, **kwargs)

    async def _process_interaction(self, payload: Dict[str, Any]):
        event_name = "dislash_interaction" if self._uses_discord_2 else "interaction"
        _type = payload.get("type", 1)
        # Received a ping
        if _type == 1:
            inter = BaseInteraction(self.client, payload)
            self.dispatch(event_name, inter)
            await inter.create_response(type=1)
        # Application command invoked
        elif _type == 2:
            data_type = payload.get("data", {}).get("type", 1)
            if data_type == 1:
                inter = SlashInteraction(self.client, payload)
                self.dispatch(event_name, inter)
                self.dispatch("slash_command", inter)
                await self._on_slash_command(inter)
            elif data_type in (2, 3):
                inter = ContextMenuInteraction(self.client, payload)
                self.dispatch(event_name, inter)
                if data_type == 2:
                    self.dispatch("user_command", inter)
                    await self._on_user_command(inter)
                elif data_type == 3:
                    self.dispatch("message_command", inter)
                    await self._on_message_command(inter)
        # Message component clicked
        elif _type == 3:
            inter = MessageInteraction(self.client, payload)
            self.dispatch(event_name, inter)
            self.dispatch("message_interaction", inter)
            if inter.component is None:
                return
            if inter.component.type == ComponentType.Button:
                self.dispatch("button_click", inter)
            elif inter.component.type == ComponentType.SelectMenu:
                self.dispatch("dropdown", inter)

    async def _fill_app_id(self):
        data = await self.client.http.application_info()
        self.application_id = int(data["id"])

    command = slash_command


SlashClient = InteractionClient



import asyncio
import datetime
import json
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import discord
from discord import AllowedMentions, Client, ClientUser, Embed, File, Member
from discord.http import Route
from discord.webhook import WebhookMessage

from .message_components import ActionRow

__all__ = ("InteractionType", "ResponseType", "BaseInteraction")


class InteractionType(int, Enum):
    Ping = 1
    ApplicationCommand = 2
    MessageComponent = 3


class ResponseType(int, Enum):
    """
    All possible response type values. Used in :class:`Interaction.reply`

    Attributes
    ----------
    Pong = 1
        ACK a Ping
    ChannelMessageWithSource = 4
        Respond to an interaction with a message
    AcknowledgeWithSource = 5
        ACK an interaction and edit a response later, the user sees a loading state
    DeferredUpdateMessage = 6
        For components, ACK an interaction and edit the original message later;
        the user does not see a loading state
    UpdateMessage = 7
        For components, edit the message the component was attached to
    """

    Pong = 1
    Acknowledge = 1
    ChannelMessage = 3
    ChannelMessageWithSource = 4
    AcknowledgeWithSource = 5
    DeferredUpdateMessage = 6
    UpdateMessage = 7


class BaseInteraction:
    """
    The base class for all interactions
    """

    def __init__(self, client: Client, data: dict):
        state = client._connection

        self.received_at = datetime.datetime.utcnow()
        self.bot: Any = client
        self.id = int(data["id"])
        self.application_id = int(data["application_id"])
        self.type: int = data["type"]
        self.token: str = data["token"]
        self.version = data["version"]

        self.guild_id: Optional[int]
        self.guild: Optional[discord.Guild]
        self.channel_id: Optional[int]
        self.channel: Optional[Union[discord.abc.GuildChannel, discord.abc.Thread, discord.abc.PrivateChannel]]
        self.author: Union[discord.User, discord.Member]
        
        if "guild_id" in data:
            self.guild_id = int(data["guild_id"])
            self.guild = client.get_guild(self.guild_id)
            assert self.guild is not None, "invalid data"
            self.author = discord.Member(data=data["member"], guild=self.guild, state=state)
        else:
            self.guild_id = None
            self.guild = None
            self.author = discord.User(state=state, data=data["user"])

        if "channel_id" in data:
            self.channel_id = int(data["channel_id"])
            self.channel = client.get_channel(self.channel_id)
        else:
            self.channel_id = None
            self.channel = None

        self._sent: bool = False

    @property
    def created_at(self):
        return datetime.datetime.utcfromtimestamp(((self.id >> 22) + 1420070400000) / 1000)

    @property
    def expired(self) -> bool:
        # In this method we're using self.received_at
        # instead of self.created_at because the IDs of all interactions
        # seem to always inherit from a timstamp which is
        # 4 seconds older than it should be
        utcnow = datetime.datetime.utcnow()
        if self._sent:
            return utcnow - self.received_at > datetime.timedelta(minutes=15)
        else:
            return utcnow - self.received_at > datetime.timedelta(seconds=3)

    @property
    def client(self) -> Client:
        return self.bot

    @property
    def me(self) -> Union[Member, ClientUser]:
        return self.guild.me if self.guild is not None else self.bot.user
    
    
    async def acknowledge(self):
        """
        ~For buttons~
        Acknowledge that the interaction was received without giving a visible response.
        Use this first if you will take more than 3 seconds to respond to an button interaction.
        """
        await self.reply(type=6)



    async def defer(self):
        """
        Creates and initial response to the interaction giving a 'bot is thinking' state.
        Use this first if you will take more than 3 seconds to respond to an interaction.
        If you use defer(), use edit() instead of reply() to make the next response.
        """
        await self.reply(type=5)


    async def reply(
        self,
        content: Any = None,
        *,
        embed: Embed = None,
        embeds: List[Embed] = None,
        components: List[ActionRow] = None,
        view=None,
        file: File = None,
        files: List[File] = None,
        tts: bool = False,
        hide_user_input: bool = False,
        ephemeral: bool = False,
        delete_after: float = None,
        allowed_mentions: AllowedMentions = None,
        type: int = None,
        fetch_response_message: bool = True,
    ):
        """
        Creates an interaction response. What's the difference between this method and
        :meth:`create_response`? If the token is no longer valid, this method sends a usual
        channel message instead of creating an interaction response. Also, this method
        fetches the interaction response message and returns it, unlike :meth:`create_response`.

        Parameters
        ----------
        content : :class:`str`
            message content
        embed : :class:`discord.Embed`
            message embed
        embeds : :class:`List[discord.Embed]`
            a list of up to 10 embeds to attach
        components : :class:`List[ActionRow]`
            a list of up to 5 action rows
        view : :class:`discord.ui.View`
            only usable with discord.py 2.0. Read more about ``View`` in
            discord.py 2.0 official documentation
        file : :class:`discord.File`
            if it's the first interaction reply, the file will be ignored due to API limitations.
            Everything else is the same as in :class:`discord.TextChannel.send()` method.
        files : List[:class:`discord.File`]
            same as ``file`` but for multiple files.
        hide_user_input : :class:`bool`
            if set to ``True``, user's input won't be displayed
        ephemeral : :class:`bool`
            if set to ``True``, your response will only be visible to the command author
        tts : :class:`bool`
            whether the message is text-to-speech or not
        delete_after : :class:`float`
            if specified, your reply will be deleted after ``delete_after`` seconds
        allowed_mentions : :class:`discord.AllowedMentions`
            controls the mentions being processed in this message.
        type : :class:`int` | :class:`ResponseType`
            sets the response type. If it's not specified, this method sets
            it according to ``hide_user_input``, ``content`` and ``embed`` params.
        fetch_response_message : :class:`bool`
            whether to fetch and return the response message. Defaults to ``True``.

        Raises
        ------
        ~discord.HTTPException
            sending the response failed
        ~discord.InvalidArgument
            Both ``embed`` and ``embeds`` are specified

        Returns
        -------
        message : :class:`discord.Message` | ``None``
            The response message that has been sent or ``None`` if the message is ephemeral
        """
        if type is None:
            is_empty_message = content is None and embed is None and embeds is None
            if is_empty_message:
                type = 1 if hide_user_input else 5
            else:
                type = 3 if hide_user_input else 4
        # Sometimes we have to use TextChannel.send() instead
        if self._sent or self.expired or type == 3:
            send_kwargs = dict(
                content=content,
                embed=embed,
                file=file,
                files=files,
                tts=tts,
                delete_after=delete_after,
                allowed_mentions=allowed_mentions,
            )
            if self.bot.slash._uses_discord_2:
                send_kwargs["view"] = view
            if self.bot.slash._modify_send:
                send_kwargs["components"] = components
            return await self.channel.send(**send_kwargs)  # type: ignore
        # Create response
        await self.create_response(
            content=content,
            type=type,
            embed=embed,
            embeds=embeds,
            components=components,
            view=view,
            ephemeral=ephemeral,
            tts=tts,
            allowed_mentions=allowed_mentions,
        )
        self._sent = True

        if view and not view.is_finished():
            self.bot._connection.store_view(view, None)

        if type == 5:
            return None

        if delete_after is not None:
            self.bot.loop.create_task(self.delete_after(delete_after))

        if fetch_response_message:
            try:
                return await self.fetch_initial_response()
            except Exception:
                pass

    async def create_response(
        self,
        content: Any = None,
        *,
        type: int = None,
        embed: Embed = None,
        embeds: List[Embed] = None,
        components: List[ActionRow] = None,
        view=None,
        ephemeral: bool = False,
        tts: bool = False,
        allowed_mentions=None,
    ):
        """
        Creates an interaction response.

        Parameters
        ----------
        content : :class:`str`
            response content
        type : :class:`int` | :class:`ResponseType`
            sets the response type. See :class:`ResponseType`
        embed : :class:`discord.Embed`
            response embed
        embeds : :class:`List[discord.Embed]`
            a list of up to 10 embeds to attach
        components : :class:`List[ActionRow]`
            a list of up to 5 action rows
        view : :class:`discord.ui.View`
            only usable with discord.py 2.0. Read more about ``View`` in
            discord.py 2.0 official documentation
        ephemeral : :class:`bool`
            if set to ``True``, your response will only be visible to the command author
        tts : :class:`bool`
            whether the message is text-to-speech or not
        allowed_mentions : :class:`discord.AllowedMentions`
            controls the mentions being processed in this message.

        Raises
        ------
        ~discord.HTTPException
            sending the response failed
        ~discord.InvalidArgument
            Both ``embed`` and ``embeds`` are specified
        """
        type = type or 4

        data: Any = {}
        if content is not None:
            data["content"] = str(content)

        if embed is not None and embeds is not None:
            raise discord.InvalidArgument("Can't pass both embed and embeds")

        if embed is not None:
            if not isinstance(embed, discord.Embed):
                raise discord.InvalidArgument("embed parameter must be discord.Embed")
            data["embeds"] = [embed.to_dict()]

        elif embeds is not None:
            if len(embeds) > 10:
                raise discord.InvalidArgument("embds parameter must be a list of up to 10 elements")
            elif not all(isinstance(embed, discord.Embed) for embed in embeds):
                raise discord.InvalidArgument("embeds parameter must be a list of discord.Embed")
            data["embeds"] = [embed.to_dict() for embed in embeds]

        if view is not None:
            if not hasattr(view, "__discord_ui_view__"):
                raise discord.InvalidArgument(f"view parameter must be View not {view.__class__!r}")

            _components = view.to_components()
        else:
            _components = None

        if components is not None:
            if len(components) > 5:
                raise discord.InvalidArgument("components must be a list of up to 5 action rows")
            _components = _components or []
            for comp in components:
                if isinstance(comp, ActionRow):
                    _components.append(comp.to_dict())
                else:
                    _components.append(ActionRow(comp).to_dict())

        if _components is not None:
            data["components"] = _components

        if content or embed or embeds:
            state = self.bot._connection
            if allowed_mentions is not None:
                if state.allowed_mentions is not None:
                    allowed_mentions = state.allowed_mentions.merge(allowed_mentions).to_dict()
                else:
                    allowed_mentions = allowed_mentions.to_dict()
                data["allowed_mentions"] = allowed_mentions

        if ephemeral:
            data["flags"] = 64
        if tts:
            data["tts"] = True

        payload: Dict[str, Any] = {"type": type}
        if data:
            payload["data"] = data

        await self.bot.http.request(
            Route("POST", "/interactions/{interaction_id}/{token}/callback", interaction_id=self.id, token=self.token),
            json=payload,
        )

    async def edit(
        self,
        content: Any = None,
        *,
        embed: Embed = None,
        embeds: List[Embed] = None,
        components: List[ActionRow] = None,
        allowed_mentions=None,
    ):
        """
        Edits the original interaction response.

        Parameters
        ----------
        content : :class:`str`
            New message content
        embed : :class:`discord.Embed`
            New message embed
        embeds : :class:`List[discord.Embed]`
            a list of up to 10 embeds of a new message
        components : :class:`List[ActionRow]`
            a list of up to 5 action rows
        allowed_mentions : :class:`discord.AllowedMentions`
            controls the mentions being processed in this message.

        Returns
        -------
        message : :class:`discord.Message`
            The message that was edited
        """
        # Form JSON params
        data: Any = {}
        if content is not None:
            data["content"] = str(content)
        # Embed or embeds
        if embed is not None and embeds is not None:
            raise discord.InvalidArgument("Can't pass both embed and embeds")

        if embed is not None:
            if not isinstance(embed, discord.Embed):
                raise discord.InvalidArgument("embed parameter must be discord.Embed")
            data["embeds"] = [embed.to_dict()]

        elif embeds is not None:
            if len(embeds) > 10:
                raise discord.InvalidArgument("embds parameter must be a list of up to 10 elements")
            elif not all(isinstance(embed, discord.Embed) for embed in embeds):
                raise discord.InvalidArgument("embeds parameter must be a list of discord.Embed")
            data["embeds"] = [embed.to_dict() for embed in embeds]
        # Maybe components
        if components is not None:
            if len(components) > 5:
                raise discord.InvalidArgument("components must be a list of up to 5 elements")
            if not all(isinstance(comp, ActionRow) for comp in components):
                raise discord.InvalidArgument("components must be a list of ActionRow")
            data["components"] = [comp.to_dict() for comp in components]
        # Allowed mentions
        state = self.bot._connection
        if allowed_mentions is not None:
            if state.allowed_mentions is not None:
                allowed_mentions = state.allowed_mentions.merge(allowed_mentions).to_dict()
            else:
                allowed_mentions = allowed_mentions.to_dict()
            data["allowed_mentions"] = allowed_mentions
        # HTTP-response
        r = await self.bot.http.request(
            Route(
                "PATCH", "/webhooks/{app_id}/{token}/messages/@original", app_id=self.application_id, token=self.token
            ),
            json=data,
        )
        return state.create_message(channel=self.channel, data=r)

    async def delete(self):
        """
        Deletes the original interaction response.
        """
        await self.bot.http.request(
            Route(
                "DELETE", "/webhooks/{app_id}/{token}/messages/@original", app_id=self.application_id, token=self.token
            )
        )

    async def delete_after(self, delay: float):
        await asyncio.sleep(delay)
        try:
            await self.delete()
        except Exception:
            pass

    async def fetch_initial_response(self) -> WebhookMessage:
        """
        Fetches the original interaction response.
        """
        data = await self.bot.http.request(
            Route("GET", "/webhooks/{app_id}/{token}/messages/@original", app_id=self.application_id, token=self.token)
        )
        state = self.bot._connection
        return state.create_message(channel=self.channel, data=data)

    async def followup(
        self,
        content: Any = None,
        *,
        embed: Embed = None,
        embeds: List[Embed] = None,
        file: File = None,
        files: List[File] = None,
        components: List[ActionRow] = None,
        view=None,
        tts: bool = False,
        ephemeral: bool = False,
        allowed_mentions=None,
        username: str = None,
        avatar_url: str = None,
    ):
        """
        Sends a followup message.

        Parameters
        ----------
        content : :class:`str`
            the followup message content
        embed : :class:`discord.Embed`
            the followup message embed
        embeds : :class:`List[discord.Embed]`
            a list of up to 10 embeds to attach
        file : :class:`discord.File`
            a file to attach to the message
        files : List[:class:`discord.File`]
            a list of files to attach to the message
        components : :class:`List[ActionRow]`
            a list of up to 5 action rows
        view : :class:`discord.ui.View`
            only usable with discord.py 2.0. Read more about ``View`` in
            discord.py 2.0 official documentation
        ephemeral : :class:`bool`
            if set to ``True``, your message will only be visible to the command author
        tts : :class:`bool`
            whether the message is text-to-speech or not
        allowed_mentions : :class:`discord.AllowedMentions`
            controls the mentions being processed in this message
        username : :class:`str`
            override the default bot name
        avatar_url : :class:`str`
            override the default avatar of the bot
        """
        route = Route(
            "POST",
            "/webhooks/{application_id}/{interaction_token}",
            application_id=self.application_id,
            interaction_token=self.token,
        )
        data: Any = {}

        if content:
            data["content"] = str(content)

        if embed is not None and embeds is not None:
            raise discord.InvalidArgument("cannot pass both embed and embeds parameter to followup()")

        if embed is not None:
            if not isinstance(embed, discord.Embed):
                raise discord.InvalidArgument("embed parameter must be discord.Embed")
            data["embeds"] = [embed.to_dict()]

        elif embeds is not None:
            if len(embeds) > 10:
                raise discord.InvalidArgument("embds parameter must be a list of up to 10 elements")
            elif not all(isinstance(embed, discord.Embed) for embed in embeds):
                raise discord.InvalidArgument("embeds parameter must be a list of discord.Embed")
            data["embeds"] = [embed.to_dict() for embed in embeds]

        if view:
            if not hasattr(view, "__discord_ui_view__"):
                raise discord.InvalidArgument(f"view parameter must be View not {view.__class__!r}")

            _components = view.to_components()
        else:
            _components = None

        if components:
            if len(components) > 5:
                raise discord.InvalidArgument("components must be a list of up to 5 action rows")
            _components = _components or []
            for comp in components:
                if isinstance(comp, ActionRow):
                    _components.append(comp.to_dict())
                else:
                    _components.append(ActionRow(comp).to_dict())

        if _components:
            data["components"] = _components

        state = self.bot._connection
        if allowed_mentions is not None:
            if state.allowed_mentions is not None:
                allowed_mentions = state.allowed_mentions.merge(allowed_mentions).to_dict()
            else:
                allowed_mentions = allowed_mentions.to_dict()
            data["allowed_mentions"] = allowed_mentions

        if ephemeral:
            data["flags"] = 64
        if tts:
            data["tts"] = True
        if username:
            data["username"] = username
        if avatar_url:
            data["avatar_url"] = avatar_url

        if file is not None and files is not None:
            raise discord.InvalidArgument("cannot pass both file and files parameter to followup()")

        if file is not None:
            files = [file]

        if files is None:
            data = await self.bot.http.request(route, json=data)
        else:
            # Send with files
            form: Any = [
                {
                    "name": "payload_json",
                    "value": json.dumps(data, separators=(",", ":"), ensure_ascii=True),
                }
            ]

            if len(files) == 1:
                file = files[0]
                form.append(
                    {
                        "name": "file",
                        "value": file.fp,
                        "filename": file.filename,
                        "content_type": "application/octet-stream",
                    }
                )
            else:
                for index, file in enumerate(files):
                    form.append(
                        {
                            "name": "file%s" % index,
                            "value": file.fp,
                            "filename": file.filename,
                            "content_type": "application/octet-stream",
                        }
                    )
            try:
                data = await self.bot.http.request(route, form=form, files=files)
            finally:
                for f in files:
                    f.close()

        msg = state.create_message(channel=self.channel, data=data)

        if view and not view.is_finished():
            message_id = None if msg is None else msg.id
            self.bot._connection.store_view(view, message_id)

        return msg

    send = reply

    respond = create_response



import asyncio
from typing import Awaitable, Callable, List

from ..interactions import ContextMenuInteraction, MessageCommand, UserCommand
from ._decohub import _HANDLER
from .core import InvokableApplicationCommand, class_name

__all__ = (
    "InvokableContextMenuCommand",
    "InvokableUserCommand",
    "InvokableMessageCommand",
    "user_command",
    "message_command",
)


class InvokableContextMenuCommand(InvokableApplicationCommand):
    def __init__(self, func: Callable[..., Awaitable], *, name=None, guild_ids: List[int] = None, **kwargs):
        super().__init__(func, name=name, **kwargs)
        self.guild_ids = guild_ids
        self._cog_class_name = class_name(func)
        self._cog_name = None
        self._cog = None

    def _inject_cog(self, cog):
        self._cog = cog
        self._cog_name = cog.qualified_name

    async def invoke(self, interaction: ContextMenuInteraction):
        if interaction.data.type == 2:
            interaction.user_command = self
        elif interaction.data.type == 3:
            interaction.message_command = self
        try:
            self._prepare_cooldowns(interaction)
            await self._run_checks(interaction)
            await self._maybe_cog_call(self._cog, interaction)
        except Exception as err:
            self._dispatch_error(self._cog, interaction, err)
            raise err


class InvokableUserCommand(InvokableContextMenuCommand):
    def __init__(self, func: Callable[..., Awaitable], *, name=None, guild_ids: List[int] = None, **kwargs):
        super().__init__(func, name=name, guild_ids=guild_ids, **kwargs)
        self.registerable = UserCommand(name=self.name)


class InvokableMessageCommand(InvokableContextMenuCommand):
    def __init__(self, func: Callable[..., Awaitable], *, name=None, guild_ids: List[int] = None, **kwargs):
        super().__init__(func, name=name, guild_ids=guild_ids, **kwargs)
        self.registerable = MessageCommand(name=self.name)


def user_command(*args, **kwargs):
    """
    A decorator that allows to build a user command, visible in a context menu.

    Parameters
    ----------
    name : :class:`str`
        name of the user command you want to respond to (equals to function name by default).
    guild_ids : :class:`List[int]`
        if specified, the client will register the command in these guilds.
        Otherwise this command will be registered globally.
    """

    def decorator(func):
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f"<{func.__qualname__}> must be a coroutine function")
        new_func = InvokableUserCommand(func, **kwargs)
        _HANDLER.user_commands[new_func.name] = new_func
        return new_func

    return decorator


def message_command(*args, **kwargs):
    """
    A decorator that allows to build a message command, visible in a context menu.

    Parameters
    ----------
    name : :class:`str`
        name of the user command you want to respond to (equals to function name by default).
    guild_ids : :class:`List[int]`
        if specified, the client will register the command in these guilds.
        Otherwise this command will be registered globally.
    """

    def decorator(func):
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f"<{func.__qualname__}> must be a coroutine function")
        new_func = InvokableMessageCommand(func, **kwargs)
        _HANDLER.message_commands[new_func.name] = new_func
        return new_func

    return decorator



from discord import DiscordException


class TypeError(DiscordException):
    """
    The base exception type for all slash-command related errors.

    This inherits from :exc:`discord.DiscordException`.

    This exception and exceptions inherited from it are handled
    in a special way as they are caught and passed into a special event
    from :class:`.SlashClient`, :func:`on_slash_command_error`.
    """

    def __init__(self, message=None, *args):
        if message is not None:
            # clean-up @everyone and @here mentions
            m = message.replace("@everyone", "@\u200beveryone").replace("@here", "@\u200bhere")
            super().__init__(m, *args)
        else:
            super().__init__(*args)


class BadArgument(TypeError):
    pass


class InteractionCheckFailure(TypeError):
    pass


class CheckAnyFailure(InteractionCheckFailure):
    def __init__(self, checks, errors):
        self.checks = checks
        self.errors = errors
        super().__init__("You do not have permission to run this command.")


class PrivateMessageOnly(InteractionCheckFailure):
    def __init__(self, message=None):
        super().__init__(message or "This command can only be used in private messages.")


class NoPrivateMessage(InteractionCheckFailure):
    def __init__(self, message=None):
        super().__init__(message or "This command cannot be used in private messages.")


class NotOwner(InteractionCheckFailure):
    pass


class CommandOnCooldown(InteractionCheckFailure):
    """Exception raised when the 
    
   lication command being invoked is on cooldown.

    This inherits from `ApplicationCommandError`

    ## Attributes

    `cooldown`: `Cooldown` (a class with attributes `rate`, `per`, and `type`)

    `retry_after`: `float` (the amount of seconds to wait before you can retry again)
    """

    def __init__(self, cooldown, retry_after):
        self.cooldown = cooldown
        self.retry_after = retry_after
        super().__init__("You are on cooldown. Try again in {:.2f}s".format(retry_after))


class NotGuildOwner(TypeError):
    pass


class MissingRole(InteractionCheckFailure):
    def __init__(self, missing_role):
        self.missing_role = missing_role
        message = "Role {0!r} is required to run this command.".format(missing_role)
        super().__init__(message)


class BotMissingRole(InteractionCheckFailure):
    def __init__(self, missing_role):
        self.missing_role = missing_role
        message = "Bot requires the role {0!r} to run this command".format(missing_role)
        super().__init__(message)


class MissingAnyRole(InteractionCheckFailure):
    def __init__(self, missing_roles):
        self.missing_roles = missing_roles

        missing = ["'{}'".format(role) for role in missing_roles]

        if len(missing) > 2:
            fmt = "{}, or {}".format(", ".join(missing[:-1]), missing[-1])
        else:
            fmt = " or ".join(missing)

        message = "You are missing at least one of the required roles: {}".format(fmt)
        super().__init__(message)


class BotMissingAnyRole(InteractionCheckFailure):
    def __init__(self, missing_roles):
        self.missing_roles = missing_roles

        missing = ["'{}'".format(role) for role in missing_roles]

        if len(missing) > 2:
            fmt = "{}, or {}".format(", ".join(missing[:-1]), missing[-1])
        else:
            fmt = " or ".join(missing)

        message = "Bot is missing at least one of the required roles: {}".format(fmt)
        super().__init__(message)


class NSFWChannelRequired(InteractionCheckFailure):
    def __init__(self, channel):
        self.channel = channel
        super().__init__("Channel '{}' needs to be NSFW for this command to work.".format(channel))


class MissingPermissions(InteractionCheckFailure):
    def __init__(self, missing_perms, *args):
        self.missing_perms = missing_perms

        missing = [perm.replace("_", " ").replace("guild", "server").title() for perm in missing_perms]

        if len(missing) > 2:
            fmt = "{}, and {}".format(", ".join(missing[:-1]), missing[-1])
        else:
            fmt = " and ".join(missing)
        message = "You are missing {} permission(s) to run this command.".format(fmt)
        super().__init__(message, *args)


class BotMissingPermissions(InteractionCheckFailure):
    def __init__(self, missing_perms, *args):
        self.missing_perms = missing_perms

        missing = [perm.replace("_", " ").replace("guild", "server").title() for perm in missing_perms]

        if len(missing) > 2:
            fmt = "{}, and {}".format(", ".join(missing[:-1]), missing[-1])
        else:
            fmt = " and ".join(missing)
        message = "Bot requires {} permission(s) to run this command.".format(fmt)
        super().__init__(message, *args)


SlashCommandError = TypeError 
SlashCheckFailure = InteractionCheckFailure



# Creating a global accumulator for decorated functions
# Also avoiding circular import
# there's no saving this, just annotate everything as Any and leave it be
from typing import Any, Dict
from discord import Client


class _HANDLER:
    client: Client = None  # type: ignore
    slash_commands: Dict[str, Any] = {}
    user_commands: Dict[str, Any] = {}
    message_commands: Dict[str, Any] = {}



from __future__ import annotations
import re
from abc import ABC, abstractmethod
from typing import cast, Iterable, List, Optional, Union

import discord
from discord import PartialEmoji

from .types import ActionRowPayload, ComponentPayload, SelectMenuPayload, SelectOptionPayload, ButtonStyle, ButtonPayload, ComponentType

__all__ = (
    "auto_rows",
    "ComponentType",
    "ButtonStyle",
    "Component",
    "Button",
    "ActionRow",
    "SelectOption",
    "MenuOption",
    "SelectMenu",
    "_component_factory",
)


ID_SOURCE = 0
MAX_ID = 25


def _partial_emoji_converter(argument: str) -> discord.PartialEmoji:
    if len(argument) < 5:
        # Sometimes unicode emojis are actually more than 1 symbol
        return discord.PartialEmoji(name=argument)

    match = re.match(r"<(a?):([a-zA-Z0-9\_]+):([0-9]+)>$", argument)
    if match is not None:
        emoji_animated = bool(match.group(1))
        emoji_name = match.group(2)
        emoji_id = int(match.group(3))

        return discord.PartialEmoji(name=emoji_name, animated=emoji_animated, id=emoji_id)

    raise discord.InvalidArgument(f"Failed to convert {argument} to PartialEmoji")


def _component_factory(data: ComponentPayload) -> Component:
    _type = data.get("type")
    if _type == 1:
        data = cast(ActionRowPayload, data)
        return ActionRow.from_dict(data)
    elif _type == 2:
        data = cast(ButtonPayload, data)
        return Button.from_dict(data)
    elif _type == 3:
        data = cast(SelectMenuPayload, data)
        return SelectMenu.from_dict(data)
    raise ValueError("Invalid component type")


def auto_rows(*buttons: Button, max_in_row: int = 5) -> List[ActionRow]:
    """
    Distributes buttons across multiple rows
    and returns the list of rows.
    Example
    -------
    ::

        rows = auto_rows(
            Button(label="Red", custom_id="red", style=4),
            Button(label="Green", custom_id="green", style=3),
            max_in_row=1
        )
        await ctx.send("Buttons", components=rows)

    Parameters
    ----------
    buttons : List[:class:`Button`]
        a list of buttons to distribute
    max_in_row : :class:`int`
        limit of buttons in a single row. Must be between 1 and 5.

    Returns
    -------
    List[:class:`ActionRow`]
        the list of rows with buttons
    """
    if not (1 <= max_in_row <= 5):
        raise discord.InvalidArgument("max_in_row parameter should be between 1 and 5.")
    return [
        ActionRow(*buttons[i : i + max_in_row]) for i
        in range(0, len(buttons), max_in_row)
    ]


class SelectOption:
    """
    This class represents an option in a select menu.

    Parameters
    ----------
    label : :class:`str`
        the user-facing name of the option, max 25 characters
    value : :class:`str`
        the dev-define value of the option, max 100 characters
    description : :class:`str`
        an additional description of the option, max 50 characters
    emoji : :class:`str`
        will add an emoji to the option
    default : :class:`bool`
        will render this option as selected by default
    """

    __slots__ = ("label", "value", "description", "emoji", "default")

    label: str
    value: str
    description: str
    emoji: PartialEmoji
    default: bool

    def __init__(
        self,
        label: str,
        value: str,
        description: Optional[str] = None,
        emoji: Optional[Union[str, PartialEmoji]] = None,
        default: bool = False,
    ) -> None:
        if isinstance(emoji, str):
            emoji = _partial_emoji_converter(emoji)

        self.label = label
        self.value = value
        self.description = description
        self.emoji = emoji
        self.default = default

    def __repr__(self) -> str:
        return (
            "<OptionSelect label={0.label!r} value={0.value!r} "
            "description={0.description!r} emoji={0.emoji!r} default={0.default!r}>"
        ).format(self)

    @classmethod
    def from_dict(cls, data: SelectOptionPayload) -> SelectOption:
        emoji = data.get("emoji")
        if emoji is not None:
            emoji = discord.PartialEmoji.from_dict(emoji)
        return cls(
            label=data["label"],
            value=data["value"],
            description=data.get("description"),
            emoji=emoji,
            default=data.get("default", False),
        )

    def to_dict(self) -> SelectOptionPayload:
        data = SelectOptionPayload(label=self.label, value=self.value)
        if self.description is not None:
            data["description"] = self.description
        if self.emoji is not None:
            data["emoji"] = self.emoji.to_dict()
        if self.default:
            data["default"] = self.default
        return data


class Component(ABC):
    """
    The base class for message components
    """

    disabled: bool
    custom_id: Optional[str]
    type: int

    def __init__(self, type: int) -> None:
        self.type = type

    @abstractmethod
    def to_dict(self) -> ComponentPayload:
        raise NotImplementedError


class SelectMenu(Component):
    """
    This class represents a select menu.

    Parameters
    ----------
    custom_id : :class:`str`
        a developer-defined identifier for the button, max 100 characters
    placeholder : :class:`str`
        custom placeholder text if nothing is selected, max 100 characters
    min_values : :class:`int`
        the minimum number of items that must be chosen; default 1, min 0, max 25
    max_values : :class:`int`
        the maximum number of items that can be chosen; default 1, max 25
    options : List[:class:`SelectOption`]
        the choices in the select, max 25
    disabled : :class:`bool`
        disable the menu, defaults to false

    Attributes
    ----------
    custom_id : :class:`str`
        a developer-defined identifier for the button, max 100 characters
    placeholder : :class:`str`
        custom placeholder text if nothing is selected, max 100 characters
    min_values : :class:`int`
        the minimum number of items that must be chosen; default 1, min 0, max 25
    max_values : :class:`int`
        the maximum number of items that can be chosen; default 1, max 25
    options : List[:class:`SelectOption`]
        the choices in the select, max 25
    disabled : :class:`bool`
        disable the menu, defaults to false
    selected_options : List[:class:`SelectOption`]
        the list of chosen options, max 25
    """

    placeholder: Optional[str] = None
    min_values: int = 1
    max_values: int = 1
    options: List[SelectOption]
    disabled: bool = False

    def __init__(
        self,
        *,
        custom_id: Optional[str] = None,
        placeholder: Optional[str] = None,
        min_values: int = 1,
        max_values: int = 1,
        options: Optional[List[SelectOption]] = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(3)
        self.custom_id = custom_id or "0"
        self.placeholder = placeholder
        self.min_values = min_values
        self.max_values = max_values
        self.options = options or []
        self.disabled = disabled
        self.selected_options: List[SelectOption] = []

    def __repr__(self) -> str:
        return (
            "<SelectMenu custom_id={0.custom_id!r} placeholder={0.placeholder!r} "
            "min_values={0.min_values!r} max_values={0.max_values!r} options={0.options!r}"
            "disabled={0.disabled} selected_options={0.selected_options!r}>"
        ).format(self)

    def _select_options(self, values: List[str]) -> None:
        self.selected_options = []
        for option in self.options:
            if option.value in values:
                self.selected_options.append(option)

    def add_option(
        self,
        label: str,
        value: str,
        description: Optional[str] = None,
        emoji: Optional[str] = None,
        default: bool = False,
    ) -> None:
        """
        Adds an option to the list of options of the menu.
        Parameters are the same as in :class:`SelectOption`.
        """
        self.options.append(
            SelectOption(
                label=label,
                value=value,
                description=description,
                emoji=emoji,
                default=default
            )
        )

    @classmethod
    def from_dict(cls, data: SelectMenuPayload) -> SelectMenu:
        return cls(
            custom_id=data.get("custom_id"),
            placeholder=data.get("placeholder"),
            min_values=data.get("min_values", 1),
            max_values=data.get("max_values", 1),
            options=[
                SelectOption.from_dict(o) for o
                in data.get("options", [])
            ],
            disabled=data.get("disabled", False),
        )

    def to_dict(self) -> SelectMenuPayload:
        payload = SelectMenuPayload(
            type=self.type,
            custom_id=self.custom_id,
            min_values=self.min_values,
            max_values=self.max_values,
            options=[o.to_dict() for o in self.options],
        )
        if self.placeholder:
            payload["placeholder"] = self.placeholder
        if self.disabled:
            payload["disabled"] = True
        return payload


class Button(Component):
    """
    Builds a button.

    Parameters
    ----------
    style : :class:`ButtonStyle`
        Style of the button
    label : :class:`str`
        Button text
    emoji : :class:`str` | :class:`discord.PartialEmoji`
        Button emoji
    custom_id : :class:`str`
        You should set it by yourself, it's not a snowflake.
        If button style is not :class:`ButtonStyle.link`, this is
        a required field
    url : :class:`str`
        If button style is :class:`ButtonStyle.link`, this is
        a required field.
    disabled : :class:`bool`
        Whether the button is disabled or not. Defaults to false.
    """

    style: ButtonStyle
    label: Optional[str]
    emoji: Optional[discord.PartialEmoji]
    url: Optional[str]

    def __init__(
        self,
        *,
        style: ButtonStyle,
        label: Optional[str] = None,
        emoji: Optional[Union[discord.PartialEmoji, str]] = None,
        custom_id: Optional[str] = None,
        url: Optional[str] = None,
        disabled: bool = False,
    ) -> None:
        global ID_SOURCE  # Ugly as hell

        if custom_id is None:
            if url is None:
                custom_id = str(ID_SOURCE)
                ID_SOURCE = (ID_SOURCE + 1) % MAX_ID
            elif style != ButtonStyle.link:
                raise discord.InvalidArgument("if you specify url, the style must be ButtonStyle.link")
        elif url is not None:
            raise discord.InvalidArgument("you can't specify both url and custom_id")
        elif style == ButtonStyle.link:
            raise discord.InvalidArgument("style 'link' expects url to be specified")

        if isinstance(emoji, str):
            emoji = _partial_emoji_converter(emoji)

        super().__init__(2)
        self.style = style
        self.label = label
        self.emoji = emoji
        self.custom_id = custom_id
        self.url = url
        self.disabled = disabled

    def __repr__(self) -> str:
        return (
            f"<Button custom_id={self.custom_id!r} label={self.label!r} "
            f"style={self.style!r} emoji={self.emoji!r} "
            f"url={self.url!r} disabled={self.disabled!r}>"
        )

    @property
    def id(self) -> Optional[str]:
        return self.custom_id

    @classmethod
    def from_dict(cls, data: ButtonPayload) -> Button:
        emoji = data.get("emoji")
        if emoji is not None:
            emoji = discord.PartialEmoji.from_dict(emoji)
        return cls(
            style=data["style"],
            label=data.get("label"),
            emoji=emoji,
            custom_id=data.get("custom_id"),
            url=data.get("url"),
            disabled=data.get("disabled", False),
        )

    def to_dict(self) -> ButtonPayload:
        payload = ButtonPayload(
            type=self.type,
            style=self.style,
        )
        if self.label is not None:
            payload["label"] = self.label
        if self.emoji is not None:
            payload["emoji"] = self.emoji.to_dict()
        if self.custom_id is not None:
            payload["custom_id"] = self.custom_id
        if self.url is not None:
            payload["url"] = self.url
        if self.disabled:
            payload["disabled"] = self.disabled
        return payload


class ActionRow(Component):
    """
    Represents an action row. Action rows are basically
    shelves for buttons.

    Parameters
    ----------
    components : :class:`List[Button]`
        a list of up to 5 buttons to place in a row
    """

    components: List[Component]

    def __init__(self, *components: Component):
        self._limit = 5
        if len(components) > self._limit:
            raise discord.InvalidArgument(f"components must be a list of up to {self._limit} elements")
        if not all(isinstance(comp, Component) for comp in components):
            raise discord.InvalidArgument("components must be a list of Component")

        super().__init__(1)
        self.components = list(components)

    def __repr__(self) -> str:
        return "<ActionRow components={0.components!r}>".format(self)

    @property
    def buttons(self) -> List[Component]:
        return self.components

    @property
    def _buttons(self) -> Iterable[Button]:
        return (c for c in self.components if isinstance(c, Button))

    @classmethod
    def from_dict(cls, data: ActionRowPayload) -> ActionRow:
        buttons = [_component_factory(elem) for elem in data.get("components", [])]
        return cls(*buttons)

    def to_dict(self) -> ActionRowPayload:
        return ActionRowPayload(
            type=self.type,
            components=[c.to_dict() for c in self.components],
        )

    def disable_buttons(self, *positions: int) -> None:
        """
        Sets ``disabled`` to ``True`` for all buttons in this row.
        """
        if not positions:
            for button in self._buttons:
                button.disabled = True
        else:
            for i in positions:
                component = self.components[i]
                if component.type == ComponentType.Button:
                    component.disabled = True

    def enable_buttons(self, *positions: int) -> None:
        """
        Sets ``disabled`` to ``False`` for all buttons in this row.
        """
        if not positions:
            for button in self._buttons:
                button.disabled = False
        else:
            for i in positions:
                component = self.components[i]
                if component.type == ComponentType.Button:
                    component.disabled = False

    def add_button(
        self,
        *,
        style: ButtonStyle,
        label: Optional[str] = None,
        emoji: Optional[str] = None,
        custom_id: Optional[str] = None,
        url: Optional[str] = None,
        disabled: bool = False,
    ) -> None:
        self.components.append(
            Button(
                style=style,
                label=label,
                emoji=emoji,
                custom_id=custom_id,
                url=url,
                disabled=disabled,
            )
        )

    def add_menu(
        self,
        *,
        custom_id: str,
        placeholder: Optional[str] = None,
        min_values: int = 1,
        max_values: int = 1,
        options: Optional[List[SelectOption]] = None,
    ) -> None:
        self.components.append(
            SelectMenu(
                custom_id=custom_id,
                placeholder=placeholder,
                min_values=min_values,
                max_values=max_values,
                options=options,
            )
        )


MenuOption = SelectOption



import asyncio
import inspect
from enum import EnumMeta
from typing import Any, Awaitable, Callable, Dict, List, Literal, Optional, Tuple, Union, get_origin

from ..interactions import (
    InteractionDataOption,
    Option,
    OptionChoice,
    SlashCommand,
    SlashInteraction,
    SlashInteractionData,
    Type,
)
from ..interactions.application_command import OptionParam
from ._decohub import _HANDLER
from .core import InvokableApplicationCommand, class_name

__all__ = ("SubCommand", "SubCommandGroup", "CommandParent", "slash_command", "command")


def fix_required(func: Callable, options: List[Option], connectors: Dict[str, str] = None) -> List[Option]:
    """Add a required to every option that doesn't have one"""
    connectors = connectors or {}

    sig = inspect.signature(func)
    for option in options:
        param = sig.parameters.get(connectors.get(option.name, option.name))
        if param is not None and not option.required and param.default is inspect.Parameter.empty:
            option.required = False
    return options


def extract_options(func: Callable) -> Tuple[List[Option], Dict[str, str]]:
    """Helper function to extract options and connectors from a function"""
    empty = inspect.Parameter.empty  # helper

    sig = inspect.signature(func)
    if "." in func.__qualname__:
        params = list(sig.parameters.values())[2:]
    else:
        params = list(sig.parameters.values())[1:]

    options = []
    connectors = {}
    for param in params:
        option_type, choices = parse_annotation(param.annotation)

        if isinstance(param.default, OptionParam):
            d = param.default
            option = Option(d.name or param.name, d.description, option_type, d.required, choices)
            if d.name is not None:
                connectors[d.name] = param.name
        else:
            option = Option(param.name, "-", type=option_type, required=param.default is empty, choices=choices)

        options.append(option)

    return options, connectors


def parse_annotation(annotation: Any) -> Tuple[int, Optional[List[OptionChoice]]]:
    """Extracts type or choices from an annotation"""
    if annotation is inspect.Parameter.empty or annotation is Any:
        return 3, None

    elif get_origin(annotation) is Literal:
        t = OptionParam.TYPES[type(annotation.__args__[0])]
        choices = [OptionChoice(str(i), i) for i in annotation.__args__]
        return t, choices

    elif isinstance(annotation, EnumMeta):
        members = [(i.name, i.value) for i in annotation]  # type: ignore
        t = OptionParam.TYPES[type(members[0][1])]
        choices = [OptionChoice(str(name).replace("_", " "), value) for name, value in members]
        return t, choices

    elif annotation in OptionParam.TYPES:
        return OptionParam.TYPES[annotation], None

    valid = ", ".join(getattr(i, "__name__", str(i)) for i in OptionParam.TYPES)
    raise TypeError(f"{annotation} is not a valid type. Must be one of: " + valid)


class BaseSlashCommand(InvokableApplicationCommand):
    def __init__(
        self, func: Callable[..., Awaitable], *, name: str = None, connectors: Dict[str, str] = None, **kwargs
    ):
        super().__init__(func, name=name, **kwargs)
        self.connectors = connectors

    def _uses_ui(self, from_cog: bool) -> bool:
        func = inspect.unwrap(self.func)
        code = func.__code__
        argcount = code.co_argcount + code.co_kwonlyargcount
        if from_cog:
            return argcount > 2
        else:
            return argcount > 1

    async def _maybe_cog_call(
        self, cog: Any, inter: SlashInteraction, data: Union[SlashInteractionData, InteractionDataOption]
    ):
        kwargs = data._to_dict_values(self.connectors) if self._uses_ui(cog) else {}
        if isinstance(self, (CommandParent, SubCommandGroup)) and self.children:
            # this fixes a bug where command parents with OptionParam recieve subcommand options
            kwargs = {}
        else:
            kwargs = self._process_arguments(inter, kwargs)

        if cog:
            return await self(cog, inter, **kwargs)
        else:
            return await self(inter, **kwargs)

    def _process_arguments(self, inter: SlashInteraction, kwargs: Dict[str, Any]):
        sig = inspect.signature(self.func)
        for param in sig.parameters.values():
            # fix accidental defaults
            if param.name not in kwargs or isinstance(kwargs[param.name], OptionParam):
                if isinstance(param.default, OptionParam):
                    if callable(param.default.default):
                        kwargs[param.name] = param.default.default(inter)
                    elif param.default.default is not ...:
                        kwargs[param.name] = param.default.default
                elif param.default is not inspect.Parameter.empty:
                    kwargs[param.name] = param.default
            elif isinstance(param.default, OptionParam) and param.default.converter is not None:
                try:
                    kwargs[param.name] = param.default.converter(inter, kwargs[param.name])
                except Exception as e:
                    raise ConversionError(param.default.converter, e) from e  # type: ignore

            # verify types
            if (
                param.name in kwargs
                and isinstance(param.default, OptionParam)
                and not self._isinstance(kwargs[param.name], param.annotation)
            ):
                error = TypeError(
                    f"Expected option {param.default.name or param.name!r} "
                    f"to be of type {param.annotation!r} but received {kwargs[param.name]!r}"
                )
                raise ConversionError(None, error) from error  # type: ignore

        return kwargs


class SubCommand(BaseSlashCommand):
    def __init__(
        self,
        func: Callable[..., Awaitable],
        *,
        name: str = None,
        description: str = None,
        options: List[Option] = None,
        connectors: Dict[str, str] = None,
        **kwargs,
    ):
        super().__init__(func, name=name, connectors=connectors, **kwargs)
        if options:
            options = fix_required(func, options)
        else:
            options, connectors = extract_options(func)
            if self.connectors:
                self.connectors.update(connectors)
            else:
                self.connectors = connectors

        self.option = Option(name=self.name, description=description or "-", type=Type.SUB_COMMAND, options=options)


class SubCommandGroup(BaseSlashCommand):
    def __init__(self, func: Callable[..., Awaitable], *, name: str = None, **kwargs):
        super().__init__(func, name=name, **kwargs)
        self.children: Dict[str, SubCommand] = {}
        self.option = Option(name=self.name, description="-", type=Type.SUB_COMMAND_GROUP, options=[])

    def sub_command(
        self,
        name: str = None,
        description: str = None,
        options: List[Option] = None,
        connectors: Dict[str, str] = None,
        **kwargs,
    ):
        """
        A decorator that creates a subcommand in the
        subcommand group.

        Parameters are the same as in :class:`CommandParent.sub_command`
        """

        def decorator(func):
            new_func = SubCommand(
                func, name=name, description=description, options=options, connectors=connectors, **kwargs
            )
            self.children[new_func.name] = new_func
            self.option.options.append(new_func.option)
            return new_func

        return decorator


class CommandParent(BaseSlashCommand):
    def __init__(
        self,
        func: Callable[..., Awaitable],
        *,
        name: str = None,
        description: str = None,
        options: List[Option] = None,
        default_permission: bool = True,
        guild_ids: List[int] = None,
        connectors: Dict[str, str] = None,
        auto_sync: bool = True,
        **kwargs,
    ):
        super().__init__(func, name=name, connectors=connectors, **kwargs)
        if options:
            options = fix_required(func, options)
        else:
            options, connectors = extract_options(func)
            if self.connectors:
                self.connectors.update(connectors)
            else:
                self.connectors = connectors

        self.children: Dict[str, Union[SubCommand, SubCommandGroup]] = {}
        self.auto_sync = auto_sync
        self.registerable = SlashCommand(
            name=self.name,
            description=description or "-",
            options=options,
            default_permission=default_permission,
        )
        self.guild_ids = guild_ids
        self.child_type: Any = None
        # Cog indication
        self._cog_class_name = class_name(func)
        self._cog_name = None
        self._cog = None

    def _inject_cog(self, cog):
        self._cog = cog
        self._cog_name = cog.qualified_name

    def sub_command(
        self,
        name: str = None,
        description: str = None,
        options: List[Option] = None,
        connectors: Dict[str, str] = None,
        **kwargs,
    ):
        """
        A decorator that creates a subcommand under the base command.

        Parameters
        ----------
        name : :class:`str`
            the name of the subcommand. Defaults to the function name
        description : :class:`str`
            the description of the subcommand
        options : :class:`list`
            the options of the subcommand for registration in API
        connectors : :class:`dict`
            which function param states for each option. If the name
            of an option already matches the corresponding function param,
            you don't have to specify the connectors. Connectors template:
            ``{"option-name": "param_name", ...}``
        """

        def decorator(func):
            if self.child_type is None:
                if len(self.registerable.options) > 0:
                    self.registerable.options = []
                self.child_type = Type.SUB_COMMAND

            new_func = SubCommand(
                func, name=name, description=description, options=options, connectors=connectors, **kwargs
            )
            self.children[new_func.name] = new_func
            self.registerable.options.append(new_func.option)
            return new_func

        return decorator

    def sub_command_group(self, name: str = None, **kwargs):
        """
        A decorator that creates a subcommand group under the base command.
        Remember that the group must have at least one subcommand.

        Parameters
        ----------
        name : :class:`str`
            the name of the subcommand group. Defaults to the function name
        """

        def decorator(func):
            if self.child_type is None:
                if len(self.registerable.options) > 0:
                    self.registerable.options = []
                self.child_type = Type.SUB_COMMAND_GROUP

            new_func = SubCommandGroup(func, name=name, **kwargs)
            self.children[new_func.name] = new_func
            self.registerable.options.append(new_func.option)
            return new_func

        return decorator

    async def invoke_children(self, interaction: SlashInteraction):
        data = interaction.data

        option = data.option_at(0)
        if option is None:
            return

        group = None
        subcmd = None
        if option.type == Type.SUB_COMMAND_GROUP:
            group = self.children.get(option.name)
        elif option.type == Type.SUB_COMMAND:
            subcmd = self.children.get(option.name)

        if group is not None:
            option = option.option_at(0)
            subcmd = None if option is None or isinstance(group, SubCommand) else group.children.get(option.name)
        if group is not None:
            interaction.invoked_with += f" {group.name}"
            interaction.sub_command_group = group
            try:
                group._prepare_cooldowns(interaction)
                await group._run_checks(interaction)
                await group._maybe_cog_call(self._cog, interaction, data)
            except Exception as err:
                group._dispatch_error(self._cog, interaction, err)
                raise err

        if subcmd is not None and option is not None:
            interaction.invoked_with += f" {subcmd.name}"
            interaction.sub_command = subcmd
            try:
                subcmd._prepare_cooldowns(interaction)
                await subcmd._run_checks(interaction)
                await subcmd._maybe_cog_call(self._cog, interaction, option)
            except Exception as err:
                subcmd._dispatch_error(self._cog, interaction, err)
                raise err

    async def invoke(self, interaction: SlashInteraction):
        interaction._wrap_choices(self.registerable)
        interaction.slash_command = self
        try:
            self._prepare_cooldowns(interaction)
            await self._run_checks(interaction)
            await self._maybe_cog_call(self._cog, interaction, interaction.data)
            await self.invoke_children(interaction)
        except Exception as err:
            self._dispatch_error(self._cog, interaction, err)
            raise err


def slash_command(*args, **kwargs) -> Callable[[Callable[..., Awaitable]], CommandParent]:
    """
    A decorator that allows to build a slash command.

    Parameters
    ----------
    auto_sync : :class:`bool`
        whether to automatically register the command or not. Defaults to ``True``
    name : :class:`str`
        name of the slash command you want to respond to (equals to function name by default).
    description : :class:`str`
        the description of the slash command. It will be visible in Discord.
    options : :class:`List[Option]`
        the list of slash command options. The options will be visible in Discord.
    default_permission : :class:`bool`
        whether the command is enabled by default when the app is added to a guild.
    guild_ids : :class:`List[int]`
        if specified, the client will register a command in these guilds.
        Otherwise this command will be registered globally.
    connectors : :class:`dict`
        which function param states for each option. If the name
        of an option already matches the corresponding function param,
        you don't have to specify the connectors. Connectors template:
        ``{"option-name": "param_name", ...}``
    """

    def decorator(func):
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f"<{func.__qualname__}> must be a coroutine function")
        new_func = CommandParent(func, **kwargs)
        _HANDLER.slash_commands[new_func.name] = new_func
        return new_func

    return decorator


command = slash_command



from .interaction import *
from .app_command_interaction import *
from .message_interaction import *
from .message_components import *
from .application_command import *
# for correct annotations
from .application_command import option_param as OptionParam
from .types import *



# type: ignore
from discord import utils, AllowedMentions, File, InvalidArgument, Message
from discord.http import Route
from discord.flags import MessageFlags

from ...interactions import ActionRow


__all__ = (
    "create_message_with_components",
    "send_with_components",
    "edit_with_components",
    "MessageWithComponents"
)


def create_message_with_components(self, *, channel, data):
    return MessageWithComponents(state=self, channel=channel, data=data)


def send_message(self, channel_id, content, *, tts=False, embed=None, nonce=None,
                 allowed_mentions=None, message_reference=None,
                 components=None):
    r = Route('POST', '/channels/{channel_id}/messages', channel_id=channel_id)
    payload = {}

    if content:
        payload['content'] = content

    if tts:
        payload['tts'] = True

    if embed:
        payload['embed'] = embed

    if nonce:
        payload['nonce'] = nonce

    if allowed_mentions:
        payload['allowed_mentions'] = allowed_mentions

    if message_reference:
        payload['message_reference'] = message_reference

    if components:
        payload['components'] = components

    return self.request(r, json=payload)


def send_files(self, channel_id, *, files, content=None, tts=False, embed=None, nonce=None,
               allowed_mentions=None, message_reference=None,
               components=None):
    r = Route('POST', '/channels/{channel_id}/messages', channel_id=channel_id)
    form = []

    payload = {'tts': tts}
    if content:
        payload['content'] = content
    if embed:
        payload['embed'] = embed
    if nonce:
        payload['nonce'] = nonce
    if allowed_mentions:
        payload['allowed_mentions'] = allowed_mentions
    if message_reference:
        payload['message_reference'] = message_reference
    if components:
        payload['components'] = components

    form.append({'name': 'payload_json', 'value': utils.to_json(payload)})
    if len(files) == 1:
        file = files[0]
        form.append({
            'name': 'file',
            'value': file.fp,
            'filename': file.filename,
            'content_type': 'application/octet-stream'
        })
    else:
        for index, file in enumerate(files):
            form.append({
                'name': 'file%s' % index,
                'value': file.fp,
                'filename': file.filename,
                'content_type': 'application/octet-stream'
            })

    return self.request(r, form=form, files=files)


async def send_with_components(messageable, content=None, *,
                               tts=False, embed=None,
                               components=None,
                               file=None, files=None,
                               delete_after=None, nonce=None,
                               allowed_mentions=None, reference=None,
                               mention_author=None):
    channel = await messageable._get_channel()
    state = messageable._state
    content = str(content) if content is not None else None
    if embed is not None:
        embed = embed.to_dict()

    if components is not None:
        if len(components) > 5:
            raise InvalidArgument("components must be a list of up to 5 action rows")
        wrapped = []
        for comp in components:
            if isinstance(comp, ActionRow):
                wrapped.append(comp)
            else:
                wrapped.append(ActionRow(comp))
        components = [comp.to_dict() for comp in wrapped]

    if allowed_mentions is None:
        allowed_mentions = state.allowed_mentions and state.allowed_mentions.to_dict()

    elif state.allowed_mentions is not None:
        allowed_mentions = state.allowed_mentions.merge(allowed_mentions).to_dict()
    else:
        allowed_mentions = allowed_mentions.to_dict()
    if mention_author is not None:
        allowed_mentions = allowed_mentions or AllowedMentions().to_dict()
        allowed_mentions['replied_user'] = bool(mention_author)

    if reference is not None:
        try:
            reference = reference.to_message_reference_dict()
        except AttributeError:
            raise InvalidArgument('reference parameter must be Message or MessageReference') from None

    if file is not None and files is not None:
        raise InvalidArgument('cannot pass both file and files parameter to send()')

    if file is not None:
        if not isinstance(file, File):
            raise InvalidArgument('file parameter must be File')

        try:
            data = await send_files(state.http, channel.id, files=[file], allowed_mentions=allowed_mentions,
                                    content=content, tts=tts, embed=embed, nonce=nonce,
                                    message_reference=reference, components=components)
        finally:
            file.close()

    elif files is not None:
        if len(files) > 10:
            raise InvalidArgument('files parameter must be a list of up to 10 elements')
        elif not all(isinstance(file, File) for file in files):
            raise InvalidArgument('files parameter must be a list of File')

        try:
            data = await send_files(state.http, channel.id, files=files, content=content, tts=tts,
                                    embed=embed, nonce=nonce, allowed_mentions=allowed_mentions,
                                    message_reference=reference, components=components)
        finally:
            for f in files:
                f.close()
    else:
        data = await send_message(state.http, channel.id, content, tts=tts, embed=embed,
                                  nonce=nonce, allowed_mentions=allowed_mentions,
                                  message_reference=reference, components=components)

    ret = state.create_message(channel=channel, data=data)
    if delete_after is not None:
        await ret.delete(delay=delete_after)
    return ret


async def edit_with_components(message, **fields):
    try:
        content = fields['content']
    except KeyError:
        pass
    else:
        if content is not None:
            fields['content'] = str(content)

    try:
        embed = fields['embed']
    except KeyError:
        pass
    else:
        if embed is not None:
            fields['embed'] = embed.to_dict()

    try:
        components = fields['components']
    except KeyError:
        pass
    else:
        if components:
            _components = []
            for comp in components:
                if isinstance(comp, ActionRow):
                    _components.append(comp.to_dict())
                else:
                    _components.append(ActionRow(comp).to_dict())
            fields['components'] = _components

    try:
        suppress = fields.pop('suppress')
    except KeyError:
        pass
    else:
        flags = MessageFlags._from_value(message.flags.value)
        flags.suppress_embeds = suppress
        fields['flags'] = flags.value

    delete_after = fields.pop('delete_after', None)

    try:
        allowed_mentions = fields.pop('allowed_mentions')
    except KeyError:
        pass
    else:
        if allowed_mentions is not None:
            if message._state.allowed_mentions is not None:
                allowed_mentions = message._state.allowed_mentions.merge(allowed_mentions).to_dict()
            else:
                allowed_mentions = allowed_mentions.to_dict()
            fields['allowed_mentions'] = allowed_mentions

    if fields:
        data = await message._state.http.edit_message(message.channel.id, message.id, **fields)
        message._update(data)

    if delete_after is not None:
        await message.delete(delay=delete_after)


class MessageWithComponents(Message):
    def __init__(self, *, state, channel, data):
        super().__init__(
            state=state,
            channel=channel,
            data=data
        )
        self.components = [ActionRow.from_dict(dat) for dat in data.get("components", [])]

    def __repr__(self):
        res = super().__repr__()
        return f"{res[:-1]} components={self.components!r}>"



from discord import (
    InvalidArgument,
    AllowedMentions,
    File,
    GuildSticker, StickerItem,
    Embed,
    Message, MessageReference, PartialMessage, MessageFlags,
    Attachment
)

from discord.ui import View
from discord.utils import MISSING

from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Union,
    overload
)

from ...interactions import ActionRow


__all__ = (
    "send",
    "edit"
)


@overload
async def send(
    self,
    content: Optional[str] = ...,
    *,
    tts: bool = ...,
    embed: Embed = ...,
    file: File = ...,
    stickers: Sequence[Union[GuildSticker, StickerItem]] = ...,
    delete_after: float = ...,
    nonce: Union[str, int] = ...,
    allowed_mentions: AllowedMentions = ...,
    reference: Union[Message, MessageReference, PartialMessage] = ...,
    mention_author: bool = ...,
    view: View = ...,
    components: list = ...,
) -> Message:
    ...


@overload
async def send(
    self,
    content: Optional[str] = ...,
    *,
    tts: bool = ...,
    embed: Embed = ...,
    files: List[File] = ...,
    stickers: Sequence[Union[GuildSticker, StickerItem]] = ...,
    delete_after: float = ...,
    nonce: Union[str, int] = ...,
    allowed_mentions: AllowedMentions = ...,
    reference: Union[Message, MessageReference, PartialMessage] = ...,
    mention_author: bool = ...,
    view: View = ...,
    components: list = ...,
) -> Message:
    ...


@overload
async def send(
    self,
    content: Optional[str] = ...,
    *,
    tts: bool = ...,
    embeds: List[Embed] = ...,
    file: File = ...,
    stickers: Sequence[Union[GuildSticker, StickerItem]] = ...,
    delete_after: float = ...,
    nonce: Union[str, int] = ...,
    allowed_mentions: AllowedMentions = ...,
    reference: Union[Message, MessageReference, PartialMessage] = ...,
    mention_author: bool = ...,
    view: View = ...,
    components: list = ...,
) -> Message:
    ...


@overload
async def send(
    self,
    content: Optional[str] = ...,
    *,
    tts: bool = ...,
    embeds: List[Embed] = ...,
    files: List[File] = ...,
    stickers: Sequence[Union[GuildSticker, StickerItem]] = ...,
    delete_after: float = ...,
    nonce: Union[str, int] = ...,
    allowed_mentions: AllowedMentions = ...,
    reference: Union[Message, MessageReference, PartialMessage] = ...,
    mention_author: bool = ...,
    view: View = ...,
    components: list = ...,
) -> Message:
    ...


async def send(
    self,
    content=None,
    *,
    tts=None,
    embed=None,
    embeds=None,
    file=None,
    files=None,
    stickers=None,
    delete_after=None,
    nonce=None,
    allowed_mentions=None,
    reference=None,
    mention_author=None,
    view=None,
    components=None,
):
    """|coro|

    Sends a message to the destination with the content given.

    The content must be a type that can convert to a string through ``str(content)``.
    If the content is set to ``None`` (the default), then the ``embed`` parameter must
    be provided.

    To upload a single file, the ``file`` parameter should be used with a
    single :class:`~discord.File` object. To upload multiple files, the ``files``
    parameter should be used with a :class:`list` of :class:`~discord.File` objects.
    **Specifying both parameters will lead to an exception**.

    To upload a single embed, the ``embed`` parameter should be used with a
    single :class:`~discord.Embed` object. To upload multiple embeds, the ``embeds``
    parameter should be used with a :class:`list` of :class:`~discord.Embed` objects.
    **Specifying both parameters will lead to an exception**.

    Parameters
    ------------
    content: Optional[:class:`str`]
        The content of the message to send.
    tts: :class:`bool`
        Indicates if the message should be sent using text-to-speech.
    embed: :class:`~discord.Embed`
        The rich embed for the content.
    file: :class:`~discord.File`
        The file to upload.
    files: List[:class:`~discord.File`]
        A list of files to upload. Must be a maximum of 10.
    nonce: :class:`int`
        The nonce to use for sending this message. If the message was successfully sent,
        then the message will have a nonce with this value.
    delete_after: :class:`float`
        If provided, the number of seconds to wait in the background
        before deleting the message we just sent. If the deletion fails,
        then it is silently ignored.
    allowed_mentions: :class:`~discord.AllowedMentions`
        Controls the mentions being processed in this message. If this is
        passed, then the object is merged with :attr:`~discord.Client.allowed_mentions`.
        The merging behaviour only overrides attributes that have been explicitly passed
        to the object, otherwise it uses the attributes set in :attr:`~discord.Client.allowed_mentions`.
        If no object is passed at all then the defaults given by :attr:`~discord.Client.allowed_mentions`
        are used instead.

        .. versionadded:: 1.4

    reference: Union[:class:`~discord.Message`, :class:`~discord.MessageReference`, :class:`~discord.PartialMessage`]
        A reference to the :class:`~discord.Message` to which you are replying, this can be created using
        :meth:`~discord.Message.to_reference` or passed directly as a :class:`~discord.Message`. You can control
        whether this mentions the author of the referenced message using the :attr:`~discord.AllowedMentions.replied_user`
        attribute of ``allowed_mentions`` or by setting ``mention_author``.

        .. versionadded:: 1.6

    mention_author: Optional[:class:`bool`]
        If set, overrides the :attr:`~discord.AllowedMentions.replied_user` attribute of ``allowed_mentions``.

        .. versionadded:: 1.6
    view: :class:`discord.ui.View`
        A Discord UI View to add to the message.
    embeds: List[:class:`~discord.Embed`]
        A list of embeds to upload. Must be a maximum of 10.

        .. versionadded:: 2.0
    stickers: Sequence[Union[:class:`~discord.GuildSticker`, :class:`~discord.StickerItem`]]
        A list of stickers to upload. Must be a maximum of 3.

        .. versionadded:: 2.0

    Raises
    --------
    ~discord.HTTPException
        Sending the message failed.
    ~discord.Forbidden
        You do not have the proper permissions to send the message.
    ~discord.InvalidArgument
        The ``files`` list is not of the appropriate size,
        you specified both ``file`` and ``files``,
        or you specified both ``embed`` and ``embeds``,
        or the ``reference`` object is not a :class:`~discord.Message`,
        :class:`~discord.MessageReference` or :class:`~discord.PartialMessage`.

    Returns
    ---------
    :class:`~discord.Message`
        The message that was sent.
    """

    channel = await self._get_channel()
    state = self._state
    content = str(content) if content is not None else None
    _components = None

    if embed is not None and embeds is not None:
        raise InvalidArgument('cannot pass both embed and embeds parameter to send()')

    if embed is not None:
        embed = embed.to_dict()

    elif embeds is not None:
        if len(embeds) > 10:
            raise InvalidArgument('embeds parameter must be a list of up to 10 elements')
        embeds = [embed.to_dict() for embed in embeds]

    if stickers is not None:
        stickers = [sticker.id for sticker in stickers]

    if allowed_mentions is None:
        allowed_mentions = state.allowed_mentions and state.allowed_mentions.to_dict()

    elif state.allowed_mentions is not None:
        allowed_mentions = state.allowed_mentions.merge(allowed_mentions).to_dict()
    else:
        allowed_mentions = allowed_mentions.to_dict()
    if mention_author is not None:
        allowed_mentions = allowed_mentions or AllowedMentions().to_dict()
        allowed_mentions['replied_user'] = bool(mention_author)

    if reference is not None:
        try:
            reference = reference.to_message_reference_dict()
        except AttributeError:
            raise InvalidArgument('reference parameter must be Message, MessageReference, or PartialMessage') from None

    if view:
        if not hasattr(view, '__discord_ui_view__'):
            raise InvalidArgument(f'view parameter must be View not {view.__class__!r}')

        _components = view.to_components()
    else:
        _components = None

    if components:
        # This is not needed, but some users still prefer
        # dislash style of sending components so yeah
        if len(components) > 5:
            raise InvalidArgument("components must be a list of up to 5 action rows")
        wrapped = []
        for comp in components:
            if isinstance(comp, ActionRow):
                wrapped.append(comp)
            else:
                wrapped.append(ActionRow(comp))
        components = _components or []
        for comp in wrapped:
            components.append(comp.to_dict())

    if file is not None and files is not None:
        raise InvalidArgument('cannot pass both file and files parameter to send()')

    if file is not None:
        if not isinstance(file, File):
            raise InvalidArgument('file parameter must be File')

        try:
            data = await state.http.send_files(
                channel.id,
                files=[file],
                allowed_mentions=allowed_mentions,
                content=content,
                tts=tts,
                embed=embed,
                embeds=embeds,
                nonce=nonce,
                message_reference=reference,
                stickers=stickers,
                components=components,
            )
        finally:
            file.close()

    elif files is not None:
        if len(files) > 10:
            raise InvalidArgument('files parameter must be a list of up to 10 elements')
        elif not all(isinstance(file, File) for file in files):
            raise InvalidArgument('files parameter must be a list of File')

        try:
            data = await state.http.send_files(
                channel.id,
                files=files,
                content=content,
                tts=tts,
                embed=embed,
                embeds=embeds,
                nonce=nonce,
                allowed_mentions=allowed_mentions,
                message_reference=reference,
                stickers=stickers,
                components=components,
            )
        finally:
            for f in files:
                f.close()
    else:
        data = await state.http.send_message(
            channel.id,
            content,
            tts=tts,
            embed=embed,
            embeds=embeds,
            nonce=nonce,
            allowed_mentions=allowed_mentions,
            message_reference=reference,
            stickers=stickers,
            components=components,
        )

    ret = state.create_message(channel=channel, data=data)
    if view:
        state.store_view(view, ret.id)

    if delete_after is not None:
        await ret.delete(delay=delete_after)
    return ret


@overload
async def edit(
    self,
    *,
    content: Optional[str] = ...,
    embed: Optional[Embed] = ...,
    attachments: List[Attachment] = ...,
    suppress: bool = ...,
    delete_after: Optional[float] = ...,
    allowed_mentions: Optional[AllowedMentions] = ...,
    view: Optional[View] = ...,
    components: Optional[list] = ...,
) -> None:
    ...


@overload
async def edit(
    self,
    *,
    content: Optional[str] = ...,
    embeds: List[Embed] = ...,
    attachments: List[Attachment] = ...,
    suppress: bool = ...,
    delete_after: Optional[float] = ...,
    allowed_mentions: Optional[AllowedMentions] = ...,
    view: Optional[View] = ...,
    components: Optional[list] = ...,
) -> None:
    ...


async def edit(
    self,
    content: Optional[str] = MISSING,
    embed: Optional[Embed] = MISSING,
    embeds: List[Embed] = MISSING,
    attachments: List[Attachment] = MISSING,
    suppress: bool = MISSING,
    delete_after: Optional[float] = None,
    allowed_mentions: Optional[AllowedMentions] = MISSING,
    view: Optional[View] = MISSING,
    components: Optional[list] = MISSING,
) -> None:
    """|coro|

    Edits the message.

    The content must be able to be transformed into a string via ``str(content)``.

    .. versionchanged:: 1.3
        The ``suppress`` keyword-only parameter was added.

    Parameters
    -----------
    content: Optional[:class:`str`]
        The new content to replace the message with.
        Could be ``None`` to remove the content.
    embed: Optional[:class:`Embed`]
        The new embed to replace the original with.
        Could be ``None`` to remove the embed.
    embeds: List[:class:`Embed`]
        The new embeds to replace the original with. Must be a maximum of 10.
        To remove all embeds ``[]`` should be passed.

        .. versionadded:: 2.0
    attachments: List[:class:`Attachment`]
        A list of attachments to keep in the message. If ``[]`` is passed
        then all attachments are removed.
    suppress: :class:`bool`
        Whether to suppress embeds for the message. This removes
        all the embeds if set to ``True``. If set to ``False``
        this brings the embeds back if they were suppressed.
        Using this parameter requires :attr:`~.Permissions.manage_messages`.
    delete_after: Optional[:class:`float`]
        If provided, the number of seconds to wait in the background
        before deleting the message we just edited. If the deletion fails,
        then it is silently ignored.
    allowed_mentions: Optional[:class:`~discord.AllowedMentions`]
        Controls the mentions being processed in this message. If this is
        passed, then the object is merged with :attr:`~discord.Client.allowed_mentions`.
        The merging behaviour only overrides attributes that have been explicitly passed
        to the object, otherwise it uses the attributes set in :attr:`~discord.Client.allowed_mentions`.
        If no object is passed at all then the defaults given by :attr:`~discord.Client.allowed_mentions`
        are used instead.

        .. versionadded:: 1.4
    view: Optional[:class:`~discord.ui.View`]
        The updated view to update this message with. If ``None`` is passed then
        the view is removed.

    Raises
    -------
    HTTPException
        Editing the message failed.
    Forbidden
        Tried to suppress a message without permissions or
        edited a message's content or embed that isn't yours.
    ~discord.InvalidArgument
        You specified both ``embed`` and ``embeds``
    """

    _components = None
    payload: Dict[str, Any] = {}
    if content is not MISSING:
        payload['content'] = str(content) if content is not None else None
    if embed is not MISSING and embeds is not MISSING:
        raise InvalidArgument('cannot pass both embed and embeds parameter to edit()')

    if embed is not MISSING:
        payload['embeds'] = [] if embed is None else [embed.to_dict()]
    elif embeds is not MISSING:
        payload['embeds'] = [e.to_dict() for e in embeds]

    if suppress is not MISSING:
        flags = MessageFlags._from_value(self.flags.value)
        flags.suppress_embeds = suppress
        payload['flags'] = flags.value

    if allowed_mentions is MISSING:
        if self._state.allowed_mentions is not None and self.author.id == self._state.self_id:
            payload['allowed_mentions'] = self._state.allowed_mentions.to_dict()
    elif allowed_mentions is not None:
        if self._state.allowed_mentions is not None:
            payload['allowed_mentions'] = self._state.allowed_mentions.merge(allowed_mentions).to_dict()
        else:
            payload['allowed_mentions'] = allowed_mentions.to_dict()

    if attachments is not MISSING:
        payload['attachments'] = [a.to_dict() for a in attachments]

    if view is not MISSING:
        self._state.prevent_view_updates_for(self.id)
        _components = view.to_components() if view else []
    if components is not MISSING:
        # Once again, this is for those who prefer dislash.py style
        if components:
            _components = _components or []
            for comp in components:
                if isinstance(comp, ActionRow):
                    _components.append(comp.to_dict())
                else:
                    _components.append(ActionRow(comp).to_dict())
        else:
            _components = _components or []

    payload["components"] = _components or []

    if payload:
        data = await self._state.http.edit_message(self.channel.id, self.id, **payload)
        self._update(data)

    if view and not view.is_finished():
        self._state.store_view(view, self.id)

    if delete_after is not None:
        await self.delete(delay=delete_after)



from setuptools import setup
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / 'README.md').read_text(encoding='utf-8')
with open('requirements.txt') as f: 
    requirements = f.read().splitlines()
with open('dislash/__init__.py', 'r') as f:
    version = [line.split('=')[1].strip(" '\"") for line in f.read().splitlines() if line.startswith('__version__')][0]


setup(
    name='disnc1.py',
    version=version,
    description='nc.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    url='https://github.com/Lordnc/disnc/',
    author='NCPD',
    author_email='nc@n811nc.com',
    keywords='python, discord, slash, commands, api, discord-api, discord-py, slash-commands, message-components, buttons, select-menus, context-menus',
    packages=['dislash', 'dislash.application_commands', 'dislash.interactions', 'dislash.application_commands._modifications'],
    python_requires='>=3.6, <4',
    install_requires=requirements,
    project_urls={
        'Documentation': 'https://amazon.com',
        'Bug Reports': 'https://amazon.com',
        'Source': 'https://github.com/Lordnc/disnc/',
    },
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
      ]
)
